import math
import logging
import torch
import torch.nn.functional as F
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from torch import nn

import torch.utils.checkpoint as checkpoint
from functools import partial
from einops import rearrange

from .pos_embed import get_3d_sincos_pos_embed, get_2d_sincos_pos_embed, get_1d_sincos_pos_embed, interpolate_pos_embed_internvideo2
from .flash_attention_class import FlashAttention

logger = logging.getLogger(__name__)

try:
    from flash_attn.modules.mlp import FusedMLP
except:
    # logger.warn(f'FusedMLP of flash_attn is not installed!!!')
    pass

try:
    from flash_attn.ops.rms_norm import DropoutAddRMSNorm
except:
    # logger.warn(f'DropoutAddRMSNorm of flash_attn is not installed!!!')
    pass


class CrossAttention(nn.Module):
    def __init__(
            self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.,
            proj_drop=0., attn_head_dim=None, out_dim=None):
        super().__init__()
        if out_dim is None:
            out_dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        if attn_head_dim is not None:
            head_dim = attn_head_dim
        all_head_dim = head_dim * self.num_heads
        self.scale = qk_scale or head_dim ** -0.5
        assert all_head_dim == dim
        
        self.q = nn.Linear(dim, all_head_dim, bias=False)
        self.k = nn.Linear(dim, all_head_dim, bias=False)
        self.v = nn.Linear(dim, all_head_dim, bias=False)
        
        if qkv_bias:
            self.q_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.k_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.v_bias = nn.Parameter(torch.zeros(all_head_dim))
        else:
            self.q_bias = None
            self.k_bias = None
            self.v_bias = None
        
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(all_head_dim, out_dim)
        self.proj_drop = nn.Dropout(proj_drop)
    
    def forward(self, x, k=None, v=None):
        B, N, C = x.shape
        N_k = k.shape[1]
        N_v = v.shape[1]
        
        q_bias, k_bias, v_bias = None, None, None
        if self.q_bias is not None:
            q_bias = self.q_bias
            k_bias = self.k_bias
            v_bias = self.v_bias
        
        q = F.linear(input=x, weight=self.q.weight, bias=q_bias)
        q = q.reshape(B, N, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)  # (B, N_head, N_q, dim)
        
        k = F.linear(input=k, weight=self.k.weight, bias=k_bias)
        k = k.reshape(B, N_k, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)
        
        v = F.linear(input=v, weight=self.v.weight, bias=v_bias)
        v = v.reshape(B, N_v, 1, self.num_heads, -1).permute(2, 0, 3, 1, 4).squeeze(0)
        
        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))  # (B, N_head, N_q, N_k)
        
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        
        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        
        return x


class AttentiveBlock(nn.Module):
    
    def __init__(self, dim, num_heads, qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., norm_layer=nn.LayerNorm, attn_head_dim=None, out_dim=None):
        super().__init__()
        
        self.norm1_q = norm_layer(dim)
        self.norm1_k = norm_layer(dim)
        self.norm1_v = norm_layer(dim)
        self.cross_attn = CrossAttention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop,
            proj_drop=drop, attn_head_dim=attn_head_dim, out_dim=out_dim)
        
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
    
    def forward(self, x_q, x_kv, pos_q, pos_k, bool_masked_pos, rel_pos_bias=None):
        x_q = self.norm1_q(x_q + pos_q)
        x_k = self.norm1_k(x_kv + pos_k)
        x_v = self.norm1_v(x_kv)
        x = self.cross_attn(x_q, k=x_k, v=x_v)
        
        return x


class AttentionPoolingBlock(AttentiveBlock):
    
    def forward(self, x):
        x_q = x.mean(1, keepdim=True)
        x_kv, pos_q, pos_k = x, 0, 0
        x = super().forward(x_q, x_kv, pos_q, pos_k, bool_masked_pos=None, rel_pos_bias=None)
        x = x.squeeze(1)
        return x


class RMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps
    
    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)


class LayerScale(nn.Module):
    def __init__(self, dim, init_values=1e-5, inplace=False, force_fp32=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))
        self.force_fp32 = force_fp32

    @torch.amp.autocast("cuda", enabled=False)
    def forward(self, x):
        if self.force_fp32:
            output_type = x.dtype
            out = x.float().mul_(self.gamma.float()) if self.inplace else x.float() * self.gamma.float()
            return out.to(dtype=output_type)
        else:
            out = x.mul_(self.gamma) if self.inplace else x * self.gamma
            return out


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0., use_flash_attn=False,
                 causal=False, norm_layer=nn.LayerNorm, qk_normalization=False, use_fused_rmsnorm=False):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5
        
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        
        self.use_flash_attn = use_flash_attn
        if use_flash_attn:
            self.causal = causal
            self.inner_attn = FlashAttention(attention_dropout=attn_drop)
        
        self.qk_normalization = qk_normalization
        self.q_norm = norm_layer(dim) if qk_normalization else nn.Identity()
        self.k_norm = norm_layer(dim) if qk_normalization else nn.Identity()
        self.use_fused_rmsnorm = use_fused_rmsnorm
    
    def _naive_attn(self, x):
        B, N, C = x.shape
        # print(x.shape, torch.cuda.memory_allocated(), torch.cuda.memory_allocated())
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)  # make torchscript happy (cannot use tensor as tuple)
        
        if self.qk_normalization:
            B_, H_, N_, D_ = q.shape
            q = self.q_norm(q.transpose(1, 2).flatten(-2, -1)).view(B_, N_, H_, D_).transpose(1, 2)
            k = self.k_norm(k.transpose(1, 2).flatten(-2, -1)).view(B_, N_, H_, D_).transpose(1, 2)
        
        attn = ((q * self.scale) @ k.transpose(-2, -1))
        # attn = attn - attn.max(-1)[0].unsqueeze(-1)  # in case of overflow for fp16
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        # print(torch.cuda.memory_allocated(), torch.cuda.memory_allocated())
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x
    
    def _flash_attn(self, x, key_padding_mask=None, need_weights=False):
        
        qkv = self.qkv(x)
        qkv = rearrange(qkv, "b s (three h d) -> b s three h d", three=3, h=self.num_heads)
        
        if self.qk_normalization:
            q, k, v = qkv.unbind(2)
            if self.use_fused_rmsnorm:
                q = self.q_norm(q.flatten(-2, -1))[0].view(q.shape)
                k = self.k_norm(k.flatten(-2, -1))[0].view(k.shape)
            else:
                q = self.q_norm(q.flatten(-2, -1)).view(q.shape)
                k = self.k_norm(k.flatten(-2, -1)).view(k.shape)
            qkv = torch.stack([q, k, v], dim=2)
        
        context, _ = self.inner_attn(
            qkv, key_padding_mask=key_padding_mask, need_weights=need_weights, causal=self.causal
        )
        outs = self.proj(rearrange(context, "b s h d -> b s (h d)"))
        outs = self.proj_drop(outs)
        return outs
    
    def forward(self, x):
        x = self._naive_attn(x) if not self.use_flash_attn else self._flash_attn(x)
        return x


class Mlp(nn.Module):
    """ MLP as used in Vision Transformer, MLP-Mixer and related networks
    """
    
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU,
                 bias=True, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)
        
        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])
    
    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)
        return x


class Block(nn.Module):
    
    def __init__(
            self, dim, num_heads, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0., init_values=None,
            drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, use_flash_attn=False, use_fused_mlp=False,
            fused_mlp_heuristic=1, with_cp=False, qk_normalization=False, layerscale_no_force_fp32=False,
            use_fused_rmsnorm=False):
        super().__init__()
        
        self.norm1 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop,
                              use_flash_attn=use_flash_attn, causal=False, norm_layer=norm_layer,
                              qk_normalization=qk_normalization,
                              use_fused_rmsnorm=use_fused_rmsnorm)
        self.ls1 = LayerScale(dim, init_values=init_values,
                              force_fp32=(not layerscale_no_force_fp32)) if init_values else nn.Identity()
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        if use_fused_mlp:
            self.mlp = FusedMLP(in_features=dim, hidden_features=mlp_hidden_dim, heuristic=fused_mlp_heuristic)
        else:
            self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        self.ls2 = LayerScale(dim, init_values=init_values,
                              force_fp32=(not layerscale_no_force_fp32)) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        
        self.with_cp = with_cp
        self.use_fused_rmsnorm = use_fused_rmsnorm
    
    def forward(self, x, residual=None):
        
        def _inner_forward(x, residual=None):
            if self.use_fused_rmsnorm:
                x, residual = self.norm1(x, residual)
                x = self.drop_path1(self.ls1(self.attn(x)))
                x, residual = self.norm2(x, residual)
                x = self.drop_path2(self.ls2(self.mlp(x)))
                return x, residual
            else:
                assert residual is None
                x = x + self.drop_path1(self.ls1(self.attn(self.norm1(x))))
                x = x + self.drop_path2(self.ls2(self.mlp(self.norm2(x))))
                return x
        
        if self.with_cp:
            # print(f"\033[31m use_checkpoint [0m")
            return checkpoint.checkpoint(_inner_forward, x, residual)
        else:
            return _inner_forward(x, residual=residual)


class PatchEmbed(nn.Module):
    """ 3D Image to Patch Embedding
    """
    
    def __init__(
            self, img_size=224, patch_size=16, in_chans=3, embed_dim=768, 
            num_frames=8, tubelet_size=1, norm_layer=None
        ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.grid_size = (
            num_frames // tubelet_size, 
            img_size[0] // patch_size[0], 
            img_size[1] // patch_size[1]
        ) # (T, H, W)
        self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        self.num_img_patches = self.grid_size[1] * self.grid_size[2]

        self.proj = nn.Conv3d(
            in_channels=in_chans, out_channels=embed_dim, 
            kernel_size=(tubelet_size, patch_size[0], patch_size[1]), 
            stride=(tubelet_size, patch_size[0], patch_size[1])
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()
    
    def forward(self, x):
        x = self.proj(x)
        x = x.flatten(3).permute(0, 2, 3, 1)  # B x C x T x HW => B x T x HW x C
        x = self.norm(x)
        return x


class Linear_Decoder(nn.Module):
    def __init__(self, in_channels=1408, out_channels=3200, 
                 norm_layer=nn.LayerNorm, clip_norm_type='l2'):
        super().__init__()
        self.clip_norm_type = clip_norm_type
        # logger.info(f'Normalization Type: {clip_norm_type}')

        self.head = nn.Linear(in_channels, out_channels)
        self.norm =  norm_layer(out_channels)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x):
        x = self.norm(self.head(x))

        if self.clip_norm_type == 'l2':
            x = x / x.norm(dim=-1, keepdim=True)
        elif self.clip_norm_type == 'none':
            pass
        else:
            raise NotImplementedError

        return x


class PretrainInternVideo2(nn.Module):
    def __init__(
            self,
            in_chans: int = 3,
            patch_size: int = 14,
            img_size: int = 224,
            qkv_bias: bool = False,
            drop_path_rate: float = 0.25,
            embed_dim: int = 1408,
            num_heads: int = 16,
            mlp_ratio: float = 48/11,
            init_values: float = 1e-5,
            qk_normalization: bool = True,
            depth: int = 40,
            use_flash_attn: bool = True,
            use_fused_rmsnorm: bool = True,
            use_fused_mlp: bool = True,
            fused_mlp_heuristic: int = 1,
            attn_pool_num_heads: int = 16,
            clip_embed_dim: int = 768,
            layerscale_no_force_fp32: bool = False,
            num_frames: int = 8,
            tubelet_size: int = 1,
            sep_pos_embed: bool = False,
            sep_image_video_pos_embed: bool = False,
            use_checkpoint: bool = False,
            checkpoint_num: int = 0,
            # for unmasked teacher
            clip_teacher_embed_dim: int = 3200,
            clip_teacher_final_dim: int = 768, # if 0, not distill final features
            clip_norm_type: str = 'l2',
            clip_return_layer: int = 1,
            clip_student_return_interval: int = 1,
        ):
        super().__init__()
        
        self.num_frames = num_frames
        self.tubelet_size = tubelet_size
        assert use_flash_attn == use_fused_rmsnorm == use_fused_mlp, 'use_flash_attn, use_fused_rmsnorm and use_fused_mlp should be consistent'
        
        self.use_flash_attn = use_flash_attn
        self.embed_dim = embed_dim

        self.depth = depth
        self.clip_norm_type = clip_norm_type
        self.return_index = []
        for i in range(clip_return_layer):
            self.return_index.append(depth - int(i * clip_student_return_interval) - 1)
        # logger.info(f'Normalization Type: {clip_norm_type}')
        # logger.info(f'Strudent Return Index: {self.return_index}')
        
        if use_fused_rmsnorm:
            norm_layer_for_blocks = partial(DropoutAddRMSNorm, eps=1e-6, prenorm=True)
        else:
            norm_layer_for_blocks = partial(RMSNorm, eps=1e-6)
        self.norm_layer_for_blocks = norm_layer_for_blocks
        self.patch_embed = PatchEmbed(
            img_size, patch_size, in_chans, embed_dim,
            num_frames=num_frames, tubelet_size=tubelet_size,
        )
        num_patches = self.patch_embed.num_patches
        num_img_patches = self.patch_embed.num_img_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        
        # stolen from https://github.com/facebookresearch/mae_st/blob/dc072aaaf640d06892e23a33b42223a994efe272/models_vit.py#L65-L73C17
        self.sep_pos_embed = sep_pos_embed
        self.sep_image_video_pos_embed = sep_image_video_pos_embed
        if sep_pos_embed:
            raise NotImplementedError
        else:
            if sep_image_video_pos_embed:
                # logger.info("Use joint position embedding, for image and video we use different pos_embed.")
                self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
                self.img_pos_embed = nn.Parameter(torch.zeros(1, num_img_patches + 1, embed_dim))
                # for CLIP decoder
                self.clip_pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
                self.clip_img_pos_embed = nn.Parameter(torch.zeros(1, num_img_patches + 1, embed_dim))
            else:
                # logger.info("Use joint position embedding, for image and video we use same pos_embed.")
                self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
                self.clip_pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]
        # choose which layer to use checkpoint
        with_cp_list = [False] * depth
        if use_checkpoint:
            for idx in range(depth):
                if idx < checkpoint_num:
                    with_cp_list[idx] = True
        # logger.info(f"Droppath rate: {dpr}")
        # logger.info(f"Checkpoint list: {with_cp_list}")
        
        self.blocks = nn.ModuleList([
            Block(embed_dim, num_heads, mlp_ratio, qkv_bias=qkv_bias,
                  norm_layer=norm_layer_for_blocks,
                  drop_path=dpr[i], init_values=init_values, attn_drop=0.,
                  use_flash_attn=use_flash_attn, use_fused_mlp=use_fused_mlp,
                  fused_mlp_heuristic=fused_mlp_heuristic,
                  with_cp=with_cp_list[i],
                  qk_normalization=qk_normalization,
                  layerscale_no_force_fp32=layerscale_no_force_fp32,
                  use_fused_rmsnorm=use_fused_rmsnorm)
            for i in range(depth)])
        self.clip_projector = AttentionPoolingBlock(
            dim=embed_dim, num_heads=attn_pool_num_heads, qkv_bias=True, qk_scale=None,
            drop=0., attn_drop=0., norm_layer=partial(nn.LayerNorm, eps=1e-5), out_dim=clip_embed_dim)
        
        # CLIP decoder
        self.clip_decoder = nn.ModuleList([
            Linear_Decoder(
                in_channels=embed_dim, 
                out_channels=clip_teacher_embed_dim, 
                norm_layer=partial(nn.LayerNorm, eps=1e-5), 
                clip_norm_type=clip_norm_type
            ) for _ in range(clip_return_layer)
        ])
        self.final_clip_decoder = nn.Identity()
        if clip_teacher_final_dim > 0:
            self.final_clip_decoder = Linear_Decoder(
                in_channels=clip_embed_dim, 
                out_channels=clip_teacher_final_dim, 
                norm_layer=partial(nn.LayerNorm, eps=1e-5), 
                clip_norm_type=clip_norm_type
            )
        
        self.init_pos_embed()
        trunc_normal_(self.cls_token, std=.02)
        self.apply(self._init_weights)
        self.fix_init_weight()

    def init_pos_embed(self):
        # logger.info("Init pos_embed from sincos pos_embed")
        if self.sep_pos_embed:
            raise NotImplementedError
        else:
            # trunc_normal_(self.pos_embed, std=.02)
            # trunc_normal_(self.clip_pos_embed, std=.02)
            pos_embed = get_3d_sincos_pos_embed(
                self.pos_embed.shape[-1], 
                self.patch_embed.grid_size[1], # height & weight
                self.patch_embed.grid_size[0], # t_size
                cls_token=True
            )
            self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))
            self.clip_pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))
            
            if self.sep_image_video_pos_embed:
                img_pos_embed = get_3d_sincos_pos_embed(
                    self.pos_embed.shape[-1], 
                    self.patch_embed.grid_size[1], # height & weight
                    1,
                    cls_token=True
                )
                self.img_pos_embed.data.copy_(torch.from_numpy(img_pos_embed).float().unsqueeze(0))
                self.clip_img_pos_embed.data.copy_(torch.from_numpy(img_pos_embed).float().unsqueeze(0))

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def fix_init_weight(self):
        def rescale(param, layer_id):
            param.div_(math.sqrt(2.0 * layer_id))

        for layer_id, layer in enumerate(self.blocks):
            rescale(layer.attn.proj.weight.data, layer_id + 1)
            rescale(layer.mlp.fc2.weight.data, layer_id + 1)
    
    @property
    def dtype(self):
        return self.patch_embed.proj.weight.dtype

    def get_num_layers(self):
        return len(self.blocks)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            'pos_embed', 
            'pos_embed_spatial', 
            'pos_embed_temporal', 
            'pos_embed_cls',
            'img_pos_embed',
            'cls_token',
            'clip_pos_embed', 
            'clip_pos_embed_spatial', 
            'clip_pos_embed_temporal', 
            'clip_pos_embed_cls',
            'clip_img_pos_embed'
        }
    
    # @torch.cuda.amp.autocast(enabled=False)
    def forward(self, x, mask=None, use_image=False, x_vis_return_idx=-1, x_vis_only=False):
        x = self.patch_embed(x.type(self.dtype))
        # print(f"x.shape: {x.shape} x.dtype: {x.dtype}, model.dtype: {self.dtype}")
        B, T, L, C = x.shape  # T: temporal; L: spatial
        x = x.view([B, T * L, C])

        # append cls token
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # add pos_embed
        if self.sep_pos_embed:
            raise NotImplementedError
        else:
            if use_image:
                if self.sep_image_video_pos_embed:
                    pos_embed = self.img_pos_embed
                else:
                    # (1, num_img_patches + 1, embed_dim)
                    # print('origin pos_embed.shape:', self.pos_embed.shape)
                    cls_pos_embed = self.pos_embed[:, 0:1, :]
                    # print('cls_pos_embed.shape:', cls_pos_embed.shape)

                    img_pos_embed = self.pos_embed[:, 1:, :].view(1, self.num_frames, self.patch_embed.num_patches // self.num_frames, self.embed_dim).mean(dim=1)
                    # print('img_pos_embed.shape:', img_pos_embed.shape)

                    pos_embed = torch.cat([cls_pos_embed, img_pos_embed], dim=1)
                    # print('final img_pos_embed.shape:', pos_embed.shape)
            else:
                pos_embed = self.pos_embed
        x = x + pos_embed

        # mask tokens, ~mask means visible
        if mask is not None:
            x = x[~mask].reshape(B, -1, C) 
        else:
            x = x.reshape(B, -1, C) 

        residual = None
        x_clip = []
        for idx, blk in enumerate(self.blocks):
            if isinstance(x, tuple) and len(x) == 2:
                x, residual = x
            # print(f"\033[31m这是{idx}, {x.shape}\033[0m")
            x = blk(x, residual=residual)
            # return intermediate features
            if idx in self.return_index:
                if isinstance(x, tuple) and len(x) == 2:
                    tmp_x, tmp_residual = x
                    if residual is not None:
                        x_clip.append(tmp_x + tmp_residual)
                else:
                    x_clip.append(x)
            if idx == (self.depth + x_vis_return_idx):
                # print(f'idx = {idx} len(self.blocks)={len(self.blocks)}')
                break

        if isinstance(x, tuple) and len(x) == 2:
            x, residual = x
            if residual is not None:
                x = x + residual
        
        x_vis = x
        if x_vis_only:
            return x_vis
        
        x_pool_vis = self.clip_projector(x_vis) 
        x_align = self.final_clip_decoder(x_pool_vis)

        # align CLIP
        x_clip = torch.stack(x_clip)
        K, B, _, C_CLIP = x_clip.shape
        # add pos_embed
        if self.sep_pos_embed: 
            raise NotImplementedError
        else:
            if use_image:
                if self.sep_image_video_pos_embed:
                    clip_pos_embed = self.clip_img_pos_embed
                else:
                    # (1, num_img_patches + 1, embed_dim)
                    # print('origin pos_embed.shape:', self.pos_embed.shape)
                    clip_cls_pos_embed = self.clip_pos_embed[:, 0:1, :]
                    # print('cls_pos_embed.shape:', cls_pos_embed.shape)

                    clip_img_pos_embed = self.clip_pos_embed[:, 1:, :].view(1, self.num_frames, self.patch_embed.num_patches // self.num_frames, self.embed_dim).mean(dim=1)
                    # print('img_pos_embed.shape:', img_pos_embed.shape)

                    clip_pos_embed = torch.cat([clip_cls_pos_embed, clip_img_pos_embed], dim=1)
                    # print('final img_pos_embed.shape:', pos_embed.shape)

            else:
                clip_pos_embed = self.clip_pos_embed
        
        clip_pos_embed = clip_pos_embed.repeat(B, 1, 1)
        if mask is not None:
            x_clip = x_clip + clip_pos_embed[~mask].view(B, -1, C_CLIP).unsqueeze(0).repeat(K, 1, 1, 1)
        else:
            x_clip = x_clip + clip_pos_embed.view(B, -1, C_CLIP).unsqueeze(0).repeat(K, 1, 1, 1)
        
        # CLIP decoder
        x_clip_align = []
        for idx, clip_decoder in enumerate(self.clip_decoder):
            x_clip_align.append(clip_decoder(x_clip[idx]))
        x_clip_align = torch.stack(x_clip_align)
        

        return x_vis, x_pool_vis, x_clip_align, x_align


def pretrain_internvideo2_1b_patch14_224(config):
    model = PretrainInternVideo2(
        in_chans=3, img_size=224, patch_size=14,
        embed_dim=1408, depth=40, num_heads=16, mlp_ratio=48/11,
        clip_embed_dim=config.vision_encoder.clip_embed_dim,
        attn_pool_num_heads=16, qkv_bias=False,
        drop_path_rate=0.25,
        init_values=0.00001,
        qk_normalization=True,
        use_flash_attn=config.vision_encoder.get('use_flash_attn', True),
        use_fused_rmsnorm=config.vision_encoder.get('use_fused_rmsnorm', True),
        use_fused_mlp=config.vision_encoder.get('use_fused_mlp', True),
        fused_mlp_heuristic=1,
        layerscale_no_force_fp32=False,
        num_frames=config.vision_encoder.num_frames,
        tubelet_size=config.vision_encoder.tubelet_size,
        sep_pos_embed=False,
        sep_image_video_pos_embed=config.vision_encoder.sep_image_video_pos_embed,
        use_checkpoint=config.vision_encoder.use_checkpoint,
        checkpoint_num=config.vision_encoder.checkpoint_num,
        clip_teacher_embed_dim=config.vision_encoder.clip_teacher_embed_dim,
        clip_teacher_final_dim=config.vision_encoder.clip_teacher_final_dim,
        clip_norm_type=config.vision_encoder.clip_norm_type,
        clip_return_layer=config.vision_encoder.clip_return_layer,
        clip_student_return_interval=config.vision_encoder.clip_student_return_interval,
    )

    if config.vision_encoder.pretrained is not None:
        # logger.info(f"Loading pretrained weights from {config.vision_encoder.pretrained}")
        # state_dict = torch.load(config.vision_encoder.pretrained, map_location='cpu')
        # state_dict = torch.load("./hf_cache/internvideo2_1b_stage2.pth")
        # another_state_dict = torch.load("./hf_cache/internvl_c_13b_224px.pth", map_location='cpu')
        # state_dict = torch.load("./hf_cache/internvideo2_1b_stage2.pth", map_location='cpu')
        state_dict = torch.load(config.vision_encoder.pretrained, map_location="cpu")

        interpolate_pos_embed_internvideo2(state_dict, model, orig_t_size=8)
        # interpolate_pos_embed_internvideo2(filtered_state_dict, model, orig_t_size=8)
        # interpolate_pos_embed_internvideo2(another_state_dict, model, orig_t_size=8)
        # import pdb; pdb.set_trace()
        filtered_state_dict = {k.replace("vision_encoder.", "", 1): v for k, v in state_dict["module"].items() if k.startswith("vision_encoder.")}
        # message = model.load_state_dict(state_dict, strict=True)
        message = model.load_state_dict(filtered_state_dict, strict=True)
        # import pdb; pdb.set_trace()
        # logger.info(message)
    else:
        logger.info("No pretrained weights!!!")
    return model


def pretrain_internvideo2_6b_patch14_224(config):
    model = PretrainInternVideo2(
        in_chans=3, img_size=224, patch_size=14,
        embed_dim=3200, depth=48, num_heads=25, mlp_ratio=4,
        clip_embed_dim=config.vision_encoder.clip_embed_dim,
        attn_pool_num_heads=16, qkv_bias=False,
        drop_path_rate=0.3,
        init_values=0.00001,
        qk_normalization=True,
        use_flash_attn=config.vision_encoder.get('use_flash_attn', True),
        use_fused_rmsnorm=config.vision_encoder.get('use_fused_rmsnorm', True),
        use_fused_mlp=config.vision_encoder.get('use_fused_mlp', True),
        fused_mlp_heuristic=1,
        layerscale_no_force_fp32=False,
        num_frames=config.vision_encoder.num_frames,
        tubelet_size=config.vision_encoder.tubelet_size,
        sep_pos_embed=False,
        sep_image_video_pos_embed=config.vision_encoder.sep_image_video_pos_embed,
        use_checkpoint=config.vision_encoder.use_checkpoint,
        checkpoint_num=config.vision_encoder.checkpoint_num,
        clip_teacher_embed_dim=config.vision_encoder.clip_teacher_embed_dim,
        clip_teacher_final_dim=config.vision_encoder.clip_teacher_final_dim,
        clip_norm_type=config.vision_encoder.clip_norm_type,
        clip_return_layer=config.vision_encoder.clip_return_layer,
        clip_student_return_interval=config.vision_encoder.clip_student_return_interval,
    )

    if config.vision_encoder.pretrained is not None:
        # logger.info(f"Loading pretrained weights from {config.vision_encoder.pretrained}")
        state_dict = torch.load(config.vision_encoder.pretrained, map_location='cpu')
        interpolate_pos_embed_internvideo2(state_dict, model, orig_t_size=8)
        msg = model.load_state_dict(state_dict, strict=False)
        # logger.info(msg)
    else:
        logger.info("No pretrained weights!!!")
    return model


if __name__ == '__main__':
    import time
    from fvcore.nn import FlopCountAnalysis
    from fvcore.nn import flop_count_table
    import numpy as np

    seed = 4217
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    num_frames = 8
    img_size = 224

    model = pretrain_internvideo2_1b_patch14_224(clip_return_layer=6).cuda().half()
    # print(model)

    # flops = FlopCountAnalysis(model, torch.rand(1, 3, num_frames, img_size, img_size).cuda().half())
    # s = time.time()
    # print(flop_count_table(flops, max_depth=1))
    # print(time.time()-s)

    mask = torch.cat([
        torch.ones(1, 8 * int(16 * 16 * 0.75)),
        torch.zeros(1, 8 * int(16 * 16 * 0.25)),
        torch.zeros(1, 1),
    ], dim=-1).to(torch.bool).cuda()
    
    output = model(torch.rand(4, 3, num_frames, img_size, img_size).cuda().half(), mask.repeat(4, 1))
    print(output[0].shape)
    print(output[1].shape)
    print(output[2].shape)
    print(output[3].shape)



from typing import List, Dict, Union
import os
import random
import tempfile
from PIL import Image, ImageSequence
import base64
import io
import re
import uuid
import json
import numpy as np
import pyarrow.fs as pf
import func_timeout
from func_timeout import func_set_timeout
import math

# fmt: on
import decord
# fmt: off


def denorm_box(points, height, width):
    new_points = []
    for p in points:
        new_points.append((round(p[0] * width), round(p[1] * height)))
    return new_points

def process_image_for_tiktok(frames: List[Image.Image], mask_boxes):
    mask_boxes = mask_boxes[:len(frames)]
    frames = [np.array(f) for f in frames]
    # assert len(mask_boxes) == len(frames)
    height, width = frames[0].shape[:2]

    new_frames = []
    for boxes, frame in zip(mask_boxes, frames):
        left, top, right, bottom = 0, 0, width, height
        for box in boxes:
            pts = np.array(denorm_box(box, height, width), np.int32)
            upper_bound = max([p[1] for p in pts]) + 30
            if bottom > upper_bound:
                bottom = upper_bound
            frame[pts[0][1]: pts[2][1], pts[0][0]: pts[1][0]] = 0
        
        new_frames.append(Image.fromarray(frame[top: bottom, left: right]))
    return new_frames

# 先将视频分成 n_frames 份。训练时，每份随机抽一帧；测试时，每份抽中间的那一帧。
def _sample_frame_indices_v2(
        total_frames: int, 
        n_frames: int, 
        is_training=False, 
        video_sampling_strategy = {},
    ):
    total_frames_idxs = list(range(total_frames))
    if total_frames <= n_frames: 
        return total_frames_idxs
    k, m = divmod(total_frames, n_frames)
    frame_splits = [total_frames_idxs[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in list(range(n_frames))]
    if is_training:
        sample_ids = [random.choice(i) for i in frame_splits]
    else:
        sample_ids = [i[(len(i)+1)//2-1] for i in frame_splits]
    return sample_ids

# 均匀抽帧，必采样首尾帧。
def _sample_frame_indices_v1(total_frames: int, n_frames: int, is_training=False, video_sampling_strategy = {}):
    if n_frames == 1:
        return [0]  # sample first frame in default
    if total_frames <= n_frames:
        return list(range(total_frames))
    sample_ids = [round(i * (total_frames - 1) / (n_frames - 1)) for i in range(n_frames)]
    return sample_ids

def conduct_disturb_frame(frame_indices):
    disturb_type = random.choice(['exchange', 'crop', 'reverse', 'discard'])
    n_frames = len(frame_indices)
    frame_indices_new = []
    if disturb_type == 'exchange':
        # 均等分成4个segments, 随机交换两个segment
        seg_len = math.ceil(n_frames / 4)
        seg_idxs = list(range(0, n_frames, seg_len))
        target_idxs = random.sample(range(0, 4), 2)
        seg_idxs[target_idxs[0]], seg_idxs[target_idxs[1]] = seg_idxs[target_idxs[1]], seg_idxs[target_idxs[0]]
        for idx in seg_idxs:
            frame_indices_new += frame_indices[idx: idx+seg_len]
    elif disturb_type == 'crop':
        # 随机截取出3/4时长，再采均匀n_frames帧
        crop_len = math.ceil(n_frames / 4)
        idx_s = random.choice(range(0, crop_len+1))
        idx_e = n_frames - 1 - (crop_len - idx_s)
        frame_indices_new = np.linspace(frame_indices[idx_s], frame_indices[idx_e], n_frames, dtype=int).tolist()
    elif disturb_type == 'reverse':
        # 随机选择长度为[1/2, 1]时长的片段进行顺序颠倒
        reverse_len = math.ceil(random.uniform(0.5,1) * n_frames)
        idx_s = random.choice(range(0, n_frames-reverse_len+1))
        idx_e = idx_s + reverse_len - 1
        frame_indices_new = frame_indices[:idx_s] + list(reversed(frame_indices[idx_s: idx_e+1])) + frame_indices[idx_e+1:]
    elif disturb_type == 'discard':
        # 随机丢弃一半帧
        frame_indices_new = random.sample(frame_indices, n_frames//2)
        frame_indices_new.sort()
    return disturb_type, frame_indices_new

@func_set_timeout(60)
def _download_file(path):
    if path.startswith("hdfs"):
        local_path = os.path.join(tempfile.gettempdir(), f'{uuid.uuid4()}_' + os.path.basename(path))

        fs = pf.HadoopFileSystem.from_uri(uri="hdfs://harunava")
        hdfs_file = fs.open_input_file(path)
        file_size = hdfs_file.size()
        if file_size > 1024 * 1024 * 1024: # 1G
            os.system(f"hadoop fs -get --ct 8 -c 512 '{path}' '{local_path}' > /dev/null 2>&1") 
        elif file_size > 1024 * 1024 * 100: # 100M
            os.system(f"hadoop fs -get '{path}' '{local_path}' > /dev/null 2>&1") 
        else:
            local_fs = pf.LocalFileSystem()
            with local_fs.open_output_stream(local_path) as local_file:
                while True:
                    chunk = hdfs_file.read(1024 * 1024 * 100)  # Reading 1MB chunks, you can adjust this as needed
                    if not chunk:
                        break
                    local_file.write(chunk)
    else:
        local_path = path

    # if not os.path.exists(local_path):
    #     raise FileNotFoundError(f'{local_path}')

    return local_path

def download_file(path):
    try:
        # with timer(f'Download {path}'):
        return _download_file(path)
    except func_timeout.exceptions.FunctionTimedOut as e:
        raise ValueError(e)
    
def load_video_from_url(video_url):
    """
    Streams video from a URL and saves it to a temporary file, 
    which can be loaded into Decord's VideoReader.
    """
    import imageio
    # Create a temporary file to store the video
    temp_video_file = tempfile.NamedTemporaryFile(delete=False, suffix=".mp4")
    temp_video_path = temp_video_file.name
    # Read the video and write it to the temporary file
    reader = imageio.get_reader(video_url, "ffmpeg")
    writer = imageio.get_writer(temp_video_path, format="ffmpeg")
    for frame in reader:
        writer.append_data(frame)
    reader.close()
    writer.close()
    return temp_video_path  # Return path for Decord to read


class VideoReader:
    def __init__(self, path: str) -> None:
        self.path = path
        self.local_path = self.preprocess()
        if self.local_path.startswith("http"):
            self.vr = decord.VideoReader(load_video_from_url(self.local_path), num_threads=1, ctx=decord.cpu(0), fault_tol=1)
        else:
            self.vr = decord.VideoReader(self.local_path, num_threads=1, ctx=decord.cpu(0), fault_tol=1)
        self.vr.seek(0)
        self._length = len(self.vr)
        self._fps = self.vr.get_avg_fps()
    
    @property
    def length(self):
        return self._length
    
    @property
    def fps(self):
        return self._fps

    def sample(self, frame_indices) -> List[Image.Image]:
        frames = self.vr.get_batch(frame_indices)
        # Ensure compatibility: Convert to NumPy array if necessary
        if hasattr(frames, "asnumpy"):  # It's a Tensor
            frames = frames.asnumpy()
        elif hasattr(frames, "numpy"):  # Older decord versions
            frames = frames.numpy()
        
        frames = [Image.fromarray(f).convert('RGB') for f in frames]
        return frames

    def preprocess(self):
        return download_file(self.path)

    def postprocess(self):
        if self.path.startswith("hdfs"):
            os.remove(self.local_path)

class ImageSeqReader:
    def __init__(self, path: List[str]) -> None:
        self.path = path
        self.local_path = self.preprocess()
        self._length = len(self.local_path)
        self._fps = None
    
    @property
    def length(self):
        return self._length
    
    @property
    def fps(self):
        return self._fps

    def sample(self, frame_indices):
        return [read_image(self.local_path[i]) for i in frame_indices]

    def preprocess(self):
        local_paths = []
        for p in self.path:
             local_paths.append(p)
        return local_paths

    def postprocess(self):
        pass

class GIFReader:
    def __init__(self, path: str) -> None:
        self.path = path
        self.local_path = self.preprocess()
        self.gif = Image.open(self.local_path)
        self._length = self.gif.n_frames
        duration = self.gif.info.get('duration', 0) / 1000  # 转换为秒
        if duration > 0:
            self._fps = 1 / duration
        else:
            self._fps = None
    
    @property
    def length(self):
        return self._length
    
    @property
    def fps(self):
        return self._fps

    def sample(self, frame_indices):
        frames = []
        i = 0
        for frame in ImageSequence.Iterator(self.gif):
            if i in frame_indices:
                frames.append(frame.convert('RGB'))
            i += 1
        return frames

    def preprocess(self):
        return download_file(self.path)

    def postprocess(self):
        if self.path.startswith("hdfs"):
            os.remove(self.local_path)

def check_frame_indices(frame_indices, total_frames, video_path):
    if frame_indices[-1] == total_frames:
        frame_indices[-1] = total_frames - 1

    valid_frame_indices = [i for i in frame_indices if i >= 0 and i < total_frames]

    if len(valid_frame_indices) != len(frame_indices):
        print(f'[Error] frame out of index. video_path={video_path}, frame_indices={frame_indices}, total_frames={total_frames}', flush=True)
    
    return valid_frame_indices


def sample_video(
    video_path: Union[str, List[str]], 
    frame_indices: List[int] = None, 
    start_frame:int=None, 
    end_frame:int=None, 
    n_frames:int = None,
    time_indices: List[float] = None,
    start_time:int=None,
    end_time:int=None,
    sampling_fps:float=None,
    mask_boxes=None,
    is_training:bool=False,
    video_sampling_strategy={'video_sampler_version': 'v1'},
    return_frame_ids: bool=False,
    ) -> List[Image.Image]:

    do_frame_disturb = video_sampling_strategy.get('do_frame_disturb', False)

    if isinstance(video_path, str):
        if video_path.endswith('.gif'):
            reader = GIFReader(video_path)
        else:
            reader = VideoReader(video_path)
    else:
        reader = ImageSeqReader(video_path)
    
    total_frames = reader.length
    fps = reader.fps

    if sampling_fps is not None:
        frame_indices = list(range(0, total_frames, round(fps / sampling_fps)))
        if len(frame_indices) > n_frames:
            frame_indices = None

    if time_indices is not None:
        frame_indices = [round(float(i) * fps) for i in time_indices]

    if start_time is not None and end_time is not None:
        start_frame = round(start_time * fps)
        end_frame = round(end_time * fps)   

    if frame_indices is None:
        start_frame = 0 if start_frame is None else round(start_frame)
        end_frame = total_frames - 1 if end_frame is None else round(end_frame)

        if end_frame == total_frames:
            end_frame -= 1

        if video_sampling_strategy['video_sampler_version'] == 'v1':
            # 均匀抽帧，必采样首尾帧。
            frame_indices = _sample_frame_indices_v1(end_frame - start_frame + 1, n_frames, is_training, video_sampling_strategy)
        elif video_sampling_strategy['video_sampler_version'] == 'v2':
            frame_indices = _sample_frame_indices_v2(end_frame - start_frame + 1, n_frames, is_training, video_sampling_strategy)
        else:
            raise ValueError(f"video_sampler_version={video_sampling_strategy['video_sampler_version']} must be 'v1' or 'v2'")
        frame_indices = [i + start_frame for i in frame_indices]
    
    frame_indices = check_frame_indices(frame_indices, total_frames, video_path)

    if do_frame_disturb:
        frame_disturb_type, frame_indices_new = conduct_disturb_frame(frame_indices)
        frame_indices_raw = frame_indices[:]
        frame_indices = frame_indices_new
        
    frames = reader.sample(frame_indices)
    if mask_boxes is not None:
        frames = process_image_for_tiktok(frames, mask_boxes)
    
    n = video_sampling_strategy.get('force_frames_n_divisible', 1)
    if n > 1 and len(frames) % n != 0:
        new_n = n - len(frames) % n
        frames.extend([Image.new(mode='RGB', size=frames[-1].size) for _ in range(new_n)])
            
    reader.postprocess()

    if do_frame_disturb:
        return frames, {"frame_indices": frame_indices, "disturb_type": frame_disturb_type, "frame_indices_raw": frame_indices_raw}
    if return_frame_ids:
        return frames, frame_indices
    return frames



def load_image_from_base64String(img_path):
    img = base64.b64decode(open(img_path, "rb").read())
    buf = io.BytesIO(img)
    img = Image.open(buf)
    return img

def read_image(image_path):
    local_file = download_file(image_path)

    if local_file.endswith('.dat'):
        image = load_image_from_base64String(local_file)
    else:
        image = Image.open(local_file).convert('RGB')
    if image_path.startswith("hdfs"):
        os.remove(local_file)
    return image


def adjust_bbox(text, frame):
    
    width, height = frame.size
    new_text = []
    start_idx = 0
    for match in re.finditer(r'\[(\d+(\.\d+)?,\s*)+\d+(\.\d+)?\]', text):
        coordinate_matches = re.findall(r"([0-9.]+)", match.group(0))
        xys = [float(coord) for coord in coordinate_matches]

        new_xys = []
        for i in range(len(xys)):
            p = xys[i]

            if width == height:
                pass
            
            if width > height and i % 2 != 0:
                p = xys[i] * height
                p += (width - height) // 2
                p = round(p / width, 2)
            
            if height > width and i % 2 == 0:
                p = xys[i] * width
                p += (height - width) // 2
                p = round(p / height, 2)
            
            new_xys.append(p)
        
        new_text.append(text[start_idx: match.span()[0]])
        new_text.append(str(new_xys))
        start_idx = match.span()[1]
    new_text.append(text[start_idx: ])
    text = ''.join(new_text)
        

    return text

def bbox_area(vertices, convert_format = True):
    if convert_format:
        vertices = list(zip(vertices[::2], vertices[1::2]))
    x0, y0 = vertices[0]
    x1, y1 = vertices[1]
    return abs((x1 - x0) * (y1 - y0))

def polygon_area(vertices, convert_format = True):
    if convert_format:
        vertices = list(zip(vertices[::2], vertices[1::2]))
    n = len(vertices)  # 多边形顶点的数量
    if n == 2:
        return bbox_area(vertices, convert_format=False)
    area = 0
    for i in range(n):
        x1, y1 = vertices[i]
        x2, y2 = vertices[(i + 1) % n]
        area += x1 * y2 - x2 * y1
    return abs(area) / 2

def get_text_len(text_line):
    l = 0
    for c in text_line:
        if '\u4e00' <= c <= '\u9fff':
            l += 1
        else:
            l += 0.5
    return l

def filter_ocr_polygon(response, area_threshold=0.0005):
    try:
        resp = json.loads(response)
    except:
        return response
    new_resp = []
    for coords, text_line in resp:
        area = polygon_area(coords, convert_format=True)
        text_len = get_text_len(text_line)
        if text_len == 0:
            continue
        if area / text_len < area_threshold:
            continue
        new_resp.append([coords, text_line])
    new_resp = json.dumps(new_resp, ensure_ascii=False)
    
    return new_resp

def put_pred_to_data_dict(prediction, data_dict):
    msg = data_dict['messages'][-1]
    if msg['role'] == 'assistant':
        msg['content'][-1]['text'] = prediction
    else:
        data_dict['messages'].append({
            "role": "assistant",
            "content": [{"type": "text", "text": prediction}]
        })

def get_prompt_from_data_dict(data_dict):
    prompt = ""
    for msg in data_dict['messages']:
        role = msg['role']
        assert role in {'system', 'user', 'assistant'}
        for content in msg['content']:
            if content['type'] == 'text':
                if content['text']:
                    prompt += f"[{role}]: {content['text']}"
            elif content['type'] == 'image':
                prompt += f"[{role}]: <image>"
            elif content['type'] == 'video':
                prompt += f"[{role}]: <video>"
            prompt += '\n'
    return prompt


"""Tests for fmd module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_fmd_basics():
    from ayase.modules.fmd import FMDModule
    _test_module_basics(FMDModule, "fmd")

def test_fmd_extract(video_sample):
    from ayase.modules.fmd import FMDModule
    m = FMDModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



# Lazy imports — model modules are loaded only when instantiated via get_vqascore_model().
# Model name lists are hardcoded to avoid importing heavy dependencies at module level.

from ...constants import HF_CACHE_DIR

_MODEL_NAME_TO_MODULE = {
    "clip-flant5-xxl": "clip_t5",
    "clip-flant5-xl": "clip_t5",
    "clip-flant5-xxl-no-system": "clip_t5",
    "clip-flant5-xxl-no-system-no-user": "clip_t5",
    "llava-v1.5-13b": "llava",
    "llava-v1.5-7b": "llava",
    "sharegpt4v-7b": "llava",
    "sharegpt4v-13b": "llava",
    "llava-v1.6-13b": "llava16",
    "instructblip-flant5-xxl": "instructblip",
    "instructblip-flant5-xl": "instructblip",
    "gpt-4-turbo": "gpt4v",
    "gpt-4o": "gpt4v",
    "gpt-4.1": "gpt4v",
    "llava-onevision-qwen2-7b-si": "llavaov",
    "llava-onevision-qwen2-7b-ov": "llavaov",
    "mplug-owl3-7b": "mplug",
    "paligemma-3b-mix-224": "paligemma",
    "paligemma-3b-mix-448": "paligemma",
    "paligemma-3b-mix-896": "paligemma",
    "internvl2-1b": "internvl",
    "internvl2-2b": "internvl",
    "internvl2-4b": "internvl",
    "internvl2-8b": "internvl",
    "internvl2-26b": "internvl",
    "internvl2-40b": "internvl",
    "internvl2-llama3-76b": "internvl",
    "internvl2.5-1b": "internvl",
    "internvl2.5-2b": "internvl",
    "internvl2.5-4b": "internvl",
    "internvl2.5-8b": "internvl",
    "internvl2.5-26b": "internvl",
    "internvl2.5-38b": "internvl",
    "internvl2.5-78b": "internvl",
    "internvl3-8b": "internvl",
    "internvl3-14b": "internvl",
    "internvl3-78b": "internvl",
    "internvideo2-chat-8b": "internvideo",
    "internvideo2-chat-8b-hd": "internvideo",
    "internvideo2-chat-8b-internlm": "internvideo",
    "internlmxcomposer25-7b": "internlm",
    "llama-3.2-1b": "llama32",
    "llama-3.2-3b": "llama32",
    "llama-3.2-1b-instruct": "llama32",
    "llama-3.2-3b-instruct": "llama32",
    "llama-guard-3-1b": "llama32",
    "llama-3.2-11b-vision": "llama32",
    "llama-3.2-11b-vision-instruct": "llama32",
    "llama-3.2-90b-vision": "llama32",
    "llama-3.2-90b-vision-instruct": "llama32",
    "llama-guard-3-11b-vision": "llama32",
    "molmo-72b-0924": "molmo",
    "molmo-7b-d-0924": "molmo",
    "molmo-7b-o-0924": "molmo",
    "molmoe-1b-0924": "molmo",
    "gemini-1.5-pro": "gemini",
    "gemini-1.5-flash": "gemini",
    "gemini-2.5-flash": "gemini",
    "gemini-2.5-pro": "gemini",
    "qwen2-vl-2b": "qwen2vl",
    "qwen2-vl-7b": "qwen2vl",
    "qwen2-vl-72b": "qwen2vl",
    "qwen2.5-vl-3b": "qwen2vl",
    "qwen2.5-vl-7b": "qwen2vl",
    "qwen2.5-vl-32b": "qwen2vl",
    "qwen2.5-vl-72b": "qwen2vl",
    "llava-video-7b": "llavavideo",
    "llava-video-72B": "llavavideo",
    "tarsier-recap-7b": "tarsier",
    "tarsier2-7b": "tarsier",
    "perception-lm-1b": "perceptionlm",
    "perception-lm-3b": "perceptionlm",
    "perception-lm-8b": "perceptionlm",
}

_MODULE_KEY_TO_IMPORT = {
    "clip_t5": (".clip_t5_model", "CLIPT5Model"),
    "llava": (".llava_model", "LLaVAModel"),
    "llava16": (".llava16_model", "LLaVA16Model"),
    "instructblip": (".instructblip_model", "InstructBLIPModel"),
    "gpt4v": (".gpt4v_model", "GPT4VModel"),
    "llavaov": (".llavaov_model", "LLaVAOneVisionModel"),
    "mplug": (".mplug_model", "mPLUGOwl3Model"),
    "paligemma": (".paligemma_model", "PaliGemmaModel"),
    "internvl": (".internvl_model", "InternVL2Model"),
    "internvideo": (".internvideo_model", "InternVideo2Model"),
    "internlm": (".internlm_model", "InternLMXComposer25Model"),
    "llama32": (".llama32_model", "LLaMA32VisionModel"),
    "molmo": (".molmo_model", "MOLMOVisionModel"),
    "gemini": (".gemini_model", "GeminiModel"),
    "qwen2vl": (".qwen2vl_model", "Qwen2VLModel"),
    "llavavideo": (".llavavideo_model", "LLaVAVideoModel"),
    "tarsier": (".tarsier_model", "TarsierModel"),
    "perceptionlm": (".perceptionlm_model", "PerceptionLMModel"),
}


def list_all_vqascore_models():
    return list(_MODEL_NAME_TO_MODULE.keys())


def get_vqascore_model(model_name, device='cuda', cache_dir=HF_CACHE_DIR, **kwargs):
    if model_name not in _MODEL_NAME_TO_MODULE:
        raise NotImplementedError(f"Unknown VQAScore model: {model_name}")
    module_key = _MODEL_NAME_TO_MODULE[model_name]
    mod_path, cls_attr = _MODULE_KEY_TO_IMPORT[module_key]
    import importlib
    mod = importlib.import_module(mod_path, package=__name__)
    model_cls = getattr(mod, cls_attr)
    return model_cls(model_name, device=device, cache_dir=cache_dir, **kwargs)



"""MS-SSIM (Multi-Scale Structural Similarity Index) module.

MS-SSIM is an improvement over SSIM that computes similarity at multiple scales.
It correlates better with human perception than single-scale SSIM.
Range: 0-1 (higher is better). Typically 0.9+ is good quality.

This is a full-reference metric requiring a reference video for comparison.
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, ValidationIssue, ValidationSeverity, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class MSSSIMModule(ReferenceBasedModule):
    name = "ms_ssim"
    description = "Multi-Scale SSIM perceptual similarity metric (full-reference)"
    default_config = {
        "scales": 5,  # Number of downsampling scales
        "weights": [0.0448, 0.2856, 0.3001, 0.2363, 0.1333],  # Per-scale weights (Wang et al.)
        "subsample": 1,  # Process every Nth frame
        "warning_threshold": 0.85,  # Warn if MS-SSIM < 0.85
        "device": "auto",  # "cuda", "cpu", or "auto"
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.scales = self.config.get("scales", 5)
        self.weights = self.config.get(
            "weights", [0.0448, 0.2856, 0.3001, 0.2363, 0.1333]
        )
        self.subsample = self.config.get("subsample", 1)
        self.warning_threshold = self.config.get("warning_threshold", 0.85)
        self.device_config = self.config.get("device", "auto")
        self.device = None
        self._ml_available = False
        self._ms_ssim_fn = None

    def setup(self) -> None:
        try:
            import torch
            from pytorch_msssim import ms_ssim

            self._ms_ssim_fn = ms_ssim

            # Set device
            if self.device_config == "auto":
                self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            else:
                self.device = torch.device(self.device_config)

            self._ml_available = True
            logger.info(f"MS-SSIM module initialized on {self.device}")

        except ImportError:
            logger.warning(
                "pytorch-msssim not installed. Install with: pip install pytorch-msssim"
            )
        except Exception as e:
            logger.warning(f"Failed to setup MS-SSIM: {e}")

    def compute_reference_score(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        """Compute MS-SSIM score comparing sample to reference.

        Args:
            sample_path: Path to distorted/test video
            reference_path: Path to reference/pristine video

        Returns:
            MS-SSIM score (0-1), or None if computation failed
        """
        try:
            # Open both videos
            ref_cap = cv2.VideoCapture(str(reference_path))
            dist_cap = cv2.VideoCapture(str(sample_path))

            ms_ssim_scores = []
            frame_idx = 0

            while True:
                ret_ref, ref_frame = ref_cap.read()
                ret_dist, dist_frame = dist_cap.read()

                if not ret_ref or not ret_dist:
                    break

                # Subsample frames
                if frame_idx % self.subsample != 0:
                    frame_idx += 1
                    continue

                # Convert BGR to RGB
                ref_rgb = cv2.cvtColor(ref_frame, cv2.COLOR_BGR2RGB)
                dist_rgb = cv2.cvtColor(dist_frame, cv2.COLOR_BGR2RGB)

                # Compute MS-SSIM for this frame
                try:
                    score = self._compute_ms_ssim_frame(ref_rgb, dist_rgb)
                    if score is not None:
                        ms_ssim_scores.append(score)
                except Exception as e:
                    logger.debug(f"Failed to compute MS-SSIM for frame {frame_idx}: {e}")

                frame_idx += 1

            ref_cap.release()
            dist_cap.release()

            if not ms_ssim_scores:
                return None

            return float(np.mean(ms_ssim_scores))

        except Exception as e:
            logger.warning(f"MS-SSIM computation failed: {e}")
            return None

    def _compute_ms_ssim_frame(self, ref_frame: np.ndarray, dist_frame: np.ndarray) -> Optional[float]:
        """Compute MS-SSIM for a single frame pair.

        Args:
            ref_frame: Reference frame (H, W, C) in RGB
            dist_frame: Distorted frame (H, W, C) in RGB

        Returns:
            MS-SSIM score (0-1), or None if computation failed
        """
        try:
            import torch

            # Convert to torch tensors (N, C, H, W)
            ref_tensor = (
                torch.from_numpy(ref_frame)
                .permute(2, 0, 1)
                .unsqueeze(0)
                .float()
                .to(self.device)
            )
            dist_tensor = (
                torch.from_numpy(dist_frame)
                .permute(2, 0, 1)
                .unsqueeze(0)
                .float()
                .to(self.device)
            )

            # Normalize uint8 [0, 255] → [0, 1] unconditionally
            ref_tensor = ref_tensor / 255.0
            dist_tensor = dist_tensor / 255.0

            # Compute MS-SSIM
            with torch.no_grad():
                ms_ssim_val = self._ms_ssim_fn(
                    ref_tensor,
                    dist_tensor,
                    data_range=1.0,
                    size_average=True,
                    win_size=11,
                )

            return ms_ssim_val.item()

        except Exception as e:
            logger.debug(f"MS-SSIM frame computation error: {e}")
            return None

    def process(self, sample: Sample) -> Sample:
        """Process sample with MS-SSIM metric.

        Checks for reference_path in sample metadata. If not found, skips processing.
        """
        if not self._ml_available:
            return sample

        if not sample.is_video:
            # For images, can still compute MS-SSIM
            return self._process_image(sample)

        # Check if sample has reference_path metadata
        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample

        if not isinstance(reference, Path):
            reference = Path(reference)

        if not reference.exists():
            logger.debug(f"Reference video not found: {reference}")
            return sample

        try:
            # Compute MS-SSIM score
            ms_ssim_score = self.compute_reference_score(sample.path, reference)

            if ms_ssim_score is None:
                return sample

            # Store in quality metrics
            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.ms_ssim = ms_ssim_score

            # Add validation issue if score is low
            if ms_ssim_score < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low MS-SSIM score: {ms_ssim_score:.3f}",
                        details={
                            "ms_ssim": ms_ssim_score,
                            "threshold": self.warning_threshold,
                        },
                        recommendation="Structural similarity degraded compared to reference. "
                        "Check for distortions or quality loss.",
                    )
                )

            logger.debug(f"MS-SSIM score for {sample.path.name}: {ms_ssim_score:.3f}")

        except Exception as e:
            logger.warning(f"MS-SSIM processing failed for {sample.path}: {e}")

        return sample

    def _process_image(self, sample: Sample) -> Sample:
        """Process image sample with MS-SSIM.

        Args:
            sample: Image sample

        Returns:
            Sample with MS-SSIM score
        """
        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample

        if not isinstance(reference, Path):
            reference = Path(reference)

        if not reference.exists():
            return sample

        try:
            # Load images
            ref_img = cv2.imread(str(reference))
            dist_img = cv2.imread(str(sample.path))

            if ref_img is None or dist_img is None:
                return sample

            # Convert BGR to RGB
            ref_rgb = cv2.cvtColor(ref_img, cv2.COLOR_BGR2RGB)
            dist_rgb = cv2.cvtColor(dist_img, cv2.COLOR_BGR2RGB)

            # Compute MS-SSIM
            ms_ssim_score = self._compute_ms_ssim_frame(ref_rgb, dist_rgb)

            if ms_ssim_score is None:
                return sample

            # Store in quality metrics
            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.ms_ssim = ms_ssim_score

            if ms_ssim_score < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low MS-SSIM score: {ms_ssim_score:.3f}",
                        details={
                            "ms_ssim": ms_ssim_score,
                            "threshold": self.warning_threshold,
                        },
                    )
                )

        except Exception as e:
            logger.warning(f"MS-SSIM image processing failed for {sample.path}: {e}")

        return sample



from typing import Dict, List
from PIL import Image
import random

from .utils import sample_video, read_image, adjust_bbox, filter_ocr_polygon


class VisionParser:
    def __init__(
        self,
        n_frames=8,
        max_n_frames=256,
        is_training=True,
        video_sampling_strategy={},
    ):
        self.n_frames = n_frames
        self.max_n_frames = max_n_frames
        self.is_training = is_training
        self.video_sampling_strategy = video_sampling_strategy

        # fmt: off
        self.data_temp = {
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": "Describe the image and the video."},
                        # 支持的 image 格式：
                        {"type": "image", "image": {"image_file": "/path/to/image"}},
                        {"type": "image", "image": {"video_file": "/path/to/video", "frame_indices": 0}},
                        # 支持的 video 格式：
                        {"type": "video", "video": {"video_file": "/path/to/video"}},
                        {"type": "video", "video": {"video_file": "/path/to/video", "frame_indices": [0, 1, 2]}},
                        {"type": "video", "video": {"video_file": "/path/to/video", "start_frame": 0, "end_frame": 100}},
                        {"type": "video", "video": {"video_file": "/path/to/video", "time_indices": [0, 1, 2]}},
                        {"type": "video", "video": {"video_file": "/path/to/video", "start_time": 0, "end_time": 100}},
                        {"type": "video", "video": {"image_file": ["/path/to/image"]}, "frame_indices": [0, 1, 2]},
                    ]
                },
                {
                    "role": "assistant",
                    "content": [
                        {"type": "text","text": "xxx"}
                    ]
                }
            ],
            "dataset": "LSMDC",
            "task": "video/caption"
        }
        # fmt: on
    
    def check_format(self, data_dict: Dict, image_processing_config: Dict):
        if image_processing_config.get('do_crop', False) and image_processing_config.get('has_coordinates', False):
            raise ValueError(f'do_crop and has_coordinates cannot be True at the same time!')

    """
    1. 将 messages 中的 image/video 替换成相应的 PIL.Image/List[PIL.Image]
    2. text 的特殊处理：调整 box；过滤面积太小的OCR
    """
    def transform(self, data_dict: Dict, image_processing_config: Dict = None) -> Dict:
        self.check_format(data_dict, image_processing_config)

        self.set_n_frames(data_dict)

        first_image = None # ugly! 需要调整box/过滤面积太小的OCR的数据只有图片任务

        for msg in data_dict['messages']:
            if isinstance(msg['content'], dict):
                msg['content'] = [msg['content']]
            for content in msg['content']:

                if content['type'] == 'image':
                    content['image'] = self.load_image_item(content['image'])
                    if first_image is None:
                        first_image = content['image']
                elif content['type'] == 'video':
                    video = self.load_video_item(content['video'])
                    content['video'] = video.pop('frames')
                    if video:
                        data_dict['extra_info']['frame_disturb_info'] = video.pop('video_info', {})
                elif content['type'] == 'text':
                    pass
                else:
                    raise ValueError(f"content['type']={content['type']} MUST be one of ['image', 'video', 'text']")
        for msg in data_dict['messages']:
            for content in msg['content']:
                if content['type'] == 'text':
                    self.postprocess_text(content, data_dict, image_processing_config, first_image)

        return data_dict['messages']
                
    # set n_frames for each vision item.
    def set_n_frames(self, data_dict):

        if isinstance(self.n_frames, int):
            n_frames = self.n_frames
        else:
            n_frames = random.choice(self.n_frames)
        
        assert n_frames <= self.max_n_frames

        curr_n_frames = 0
        has_dynamic = False
        for msg in data_dict['messages']:
            if isinstance(msg['content'], dict):
                msg['content'] = [msg['content']]

            for content in msg['content']:

                if content['type'] == 'image':
                    curr_n_frames += 1 
                elif content['type'] == 'video':
                    if 'frame_indices' in content['video']:                        
                        curr_n_frames += len(content['video']['frame_indices'])
                        content['video']['n_frames'] = len(content['video']['frame_indices'])
                    elif 'time_indices' in content['video']:
                        curr_n_frames += len(content['video']['time_indices'])
                        content['video']['n_frames'] = len(content['video']['time_indices'])
                    elif 'min_n_frames' in content['video']:
                        content['video']['min_n_frames'] = int(content['video']['min_n_frames'])
                        curr_n_frames += content['video']['min_n_frames']
                        content['video']['n_frames'] = content['video']['min_n_frames']
                        has_dynamic = True        
                    elif 'fps' in content['video']:
                        content['video']['n_frames'] = self.max_n_frames
                        curr_n_frames += self.max_n_frames
                        has_dynamic = True        
                    else:
                        content['video']['n_frames'] = 0
                        has_dynamic = True

        while curr_n_frames < n_frames and has_dynamic:
            for msg in data_dict['messages']:
                for content in msg['content']:
                    if content['type'] == 'video':
                        if 'frame_indices' in content['video']:
                            pass
                        elif 'time_indices' in content['video']:
                            pass
                        else:
                            if curr_n_frames < n_frames:
                                content['video']['n_frames'] += 1
                            curr_n_frames += 1
        
        while curr_n_frames > self.max_n_frames and has_dynamic:
            for msg in data_dict['messages']:
                for content in msg['content']:
                    if content['type'] == 'video':
                        if 'frame_indices' in content['video']:
                            pass
                        elif 'time_indices' in content['video']:
                            pass
                        else:
                            if curr_n_frames > self.max_n_frames:
                                content['video']['n_frames'] -= 1
                            curr_n_frames -= 1
    

        for msg in data_dict['messages']:
            for content in msg['content']:
                if content['type'] == 'video':
                    if 'frame_indices' in content['video']:
                        pass
                    elif 'time_indices' in content['video']:
                        pass
                    else:
                        n = self.video_sampling_strategy.get('force_frames_n_divisible', 1)
                        if n > 1 and content['video']['n_frames'] % n != 0:
                            content['video']['n_frames'] += n - content['video']['n_frames'] % n

    def load_image_item(self, image_item) -> Image.Image:
        """
        image_item:
        {"image_file": {"lq": "/path/to/image"}}
        {"video_file": {"lq": "/path/to/video"}, "frame_indices": 0}
        """

        # check format
        if ("image_file" not in image_item) and ("video_file" not in image_item):
            raise KeyError(f"Key 'image_file' or 'video_file' not found in image_item")
        if 'image_file' in image_item:
            if not isinstance(image_item['image_file'], str):
                raise ValueError(f"{image_item['image_file']} is not a str!")
        if 'video_file' in image_item:
            if not isinstance(image_item['frame_indices'], int):
                raise ValueError(f"{image_item['frame_indices']} is not a int!")

        if 'image_file' in image_item:
            image = read_image(image_item['image_file'])
        else:
            frame_indices = [image_item['frame_indices']]
            image = sample_video(image_item['video_file'], frame_indices = frame_indices)[0]

        return image

    def load_video_item(self, video_item) -> List[Image.Image]:
        """
        video_item:
        {"video_file": {"lq": "/path/to/video"}, "n_frames": 8} 
        {"video_file": {"lq": "/path/to/video"}, "frame_indices": [0, 1, 2], "n_frames": 3} 
        {"video_file": {"lq": "/path/to/video"}, "start_frame": 0, "end_frame": 100, "n_frames": 8}
        {"video_file": {"lq": "/path/to/video"}, "time_indices": [0, 1, 2], "n_frames": 3}
        {"video_file": {"lq": "/path/to/video"}, "start_time": 0, "end_time": 100, "n_frames": 8}
        {"image_file": {"lq": ["/path/to/image"]}, "frame_indices": [0, 1, 2], "n_frames": 3}
        """

        # check format
        if ("image_file" not in video_item) and ("video_file" not in video_item):
            raise KeyError(f"Key 'image_file' or 'video_file' not found in video_item")
    
        video_path = video_item.get('video_file', video_item.get('image_file'))
        n_frames = video_item.get('n_frames', None)
        frame_indices = video_item.get('frame_indices', None)
        start_frame = video_item.get('start_frame', None)
        end_frame = video_item.get('end_frame', None)
        time_indices = video_item.get('time_indices', None)
        start_time = video_item.get('start_time', None)
        end_time = video_item.get('end_time', None)
        mask_boxes = video_item.get('mask_boxes', None)
        fps = video_item.get('fps', None)

        frames, frame_indices = sample_video(
            video_path=video_path,
            frame_indices=frame_indices,
            start_frame=start_frame,
            end_frame=end_frame,
            n_frames=n_frames,
            time_indices=time_indices,
            start_time=start_time,
            end_time=end_time,
            sampling_fps=fps,
            mask_boxes=mask_boxes,
            is_training=self.is_training,
            video_sampling_strategy=self.video_sampling_strategy,
            return_frame_ids=True,
        )

        if self.video_sampling_strategy.get('use_multi_images_for_video', False):
            new_frames = []
            for f in frames:
                new_frames.extend([f, f])
            frames = new_frames

        if isinstance(frame_indices, dict):
            return {
                'frames': frames,
                'video_info': frame_indices
            }
        return {'frames': frames}
    
    def postprocess_text(self, content, data_dict, image_processing_config, first_image):
        if image_processing_config.get('has_coordinates') and image_processing_config.get('do_padding'):
            content['text'] = adjust_bbox(content['text'], frame=first_image)
        if data_dict.get('task') == 'image/OCR' and image_processing_config.get('has_coordinates'):
            content['text'] = filter_ocr_polygon(content['text'])



import torch
import os

import os.path as osp
from copy import deepcopy
from importlib import import_module

from ...video_utils import get_video_details, load_frames_from_video
import numpy as np

from torchvision import transforms
from torchvision.transforms import InterpolationMode

from .multi_modality.tasks.shared_utils import setup_model
from .multi_modality.models import InternVideo2_Stage2

from .multi_modality.utils.config_utils import setup_main

import datetime
import logging
import time

import numpy as np
import torch
import torch.distributed as dist
import torch.nn.functional as F
from einops import rearrange

from ...video_utils import get_video_details, load_frames_from_video

from .multi_modality.models.criterions import get_sim

logger = logging.getLogger(__name__)


def extract_text_feats(texts, max_txt_l, tokenizer, model, device, return_ids=False):
    text_feats = []
    text_atts = []

    if return_ids:
        text_ids = []

    text_input = tokenizer(
        texts,
        padding="max_length",
        truncation=True,
        max_length=max_txt_l,
        return_tensors="pt",
    ).to(device)  # NOTE not need to cast

    text_feat = model.encode_text(text_input)[0]
    text_feats.append(text_feat)
    text_atts.append(text_input.attention_mask)
    if return_ids:
        text_ids.append(text_input.input_ids)

    text_feats = torch.cat(text_feats, dim=0)
    text_atts = torch.cat(text_atts, dim=0)
    if return_ids:
        text_ids = torch.cat(text_ids, dim=0)
        return text_feats, text_atts, text_ids
    else:
        return text_feats, text_atts


def extract_vision_feats(image_paths, transforms, model, device, num_frames=4):
    cast_dtype = None

    image_feats_all = []
    pooled_image_feats_all = []

    image = []
    for data_path in image_paths:
        total_frames, original_fps, video_duration = get_video_details(data_path)
        uniform_sampling = min(num_frames, total_frames)
        all_indices = np.linspace(0, total_frames - 1, uniform_sampling, dtype=int)
        frames = load_frames_from_video(data_path, all_indices, "decord", True)
        frames = frames.permute(0, 3, 1, 2)  # (T, C, H, W), torch.uint8
        frames = transforms(frames)
        image.append(frames)
        
    image = torch.stack(image, dim=0)
    image = image.to(device, dtype=cast_dtype, non_blocking=True)
    image_feat, pooled_image_feat = model.encode_vision(image, test=True)
    # if len(pooled_image_feat.shape) == 2:
    pooled_image_feat = pooled_image_feat.unsqueeze(1)  # make av_fusion happy
    # if config.evaluation.eval_frame_ensemble == "concat":
        # if len(image_feat.shape) == 4:
        #     image_feat = rearrange(image_feat, "b t l c -> b (t l) c").contiguous()
    image_feat = image_feat.unsqueeze(1)  # (bsz, 1, #frm*L, d)
    # import pdb; pdb.set_trace()
    # else:
    #     assert config.video_input.num_frames == 1, "only support single-frame"
    #     assert config.evaluation.eval_frame_ensemble in ["mean", "max", "lse"]
    # if config.evaluation.eval_offload:
    #     image_feats_all.append(image_feat.cpu())
    #     pooled_image_feats_all.append(pooled_image_feat.cpu())
    # else:
    image_feats_all.append(image_feat)
    pooled_image_feats_all.append(pooled_image_feat)

    image_feats_all = torch.cat(image_feats_all, dim=0)
    pooled_image_feats_all = torch.cat(pooled_image_feats_all, dim=0)

    return image_feats_all, pooled_image_feats_all


def create_transforms(image_res=224):
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)
    normalize = transforms.Normalize(mean, std)
    # loaded images and videos are torch.Tensor of torch.uint8 format,
    # ordered as (T, 1 or 3, H, W) where T=1 for image
    type_transform = transforms.Lambda(lambda x: x.float().div(255.0))

    test_transform = transforms.Compose(
        [
            transforms.Resize(
                (image_res, image_res),
                interpolation=InterpolationMode.BICUBIC,
            ),
            type_transform,
            normalize,
        ]
    )
    return test_transform


def download_internvideo2(model_name, pretrained_name, cache_dir, device="cuda"):
    repo_id = f"zhiqiulin/{model_name}"
    filename = f"{model_name}.pth"
    model_path = os.path.join(cache_dir, model_name) + ".pth"
    if not os.path.exists(model_path) or os.path.getsize(model_path) < 1000:
        hf_url = f"https://huggingface.co/{repo_id}/resolve/main/{filename}"
        os.system(f"wget -O {model_path} {hf_url}")
        # if download fails raise an error
        if not os.path.exists(model_path) or os.path.getsize(model_path) < 1000:
            raise FileNotFoundError(f"Model not found at {hf_url} because the download failed")

    # pretrained_repo_id = f"zhiqiulin/{pretrained_name}"
    # pretrained_filename = f"{pretrained_name}.pth"
    # pretrained_path = os.path.join(cache_dir, pretrained_filename)
    # if not os.path.exists(pretrained_path) or os.path.getsize(pretrained_path) < 1000:
    #     hf_url = f"https://huggingface.co/{pretrained_repo_id}/resolve/main/{pretrained_filename}"
    #     os.system(f"wget -O {pretrained_path} {hf_url}")
    #     # if download fails raise an error
    #     if not os.path.exists(pretrained_path) or os.path.getsize(pretrained_path) < 1000:
    #         raise FileNotFoundError(f"Model not found at {hf_url} because the download failed")

    current_dir = os.path.dirname(os.path.abspath(__file__))  # Get the current script's directory
    config_file = os.path.join(current_dir, "configs.py")  # Construct the new file path
    config = setup_main(config_file=config_file, pretrained_path=model_path)

    # is_pretrain = config.mode == "pt"
    # device = torch.device(config.device)

    # train_loaders, test_name2loaders, train_media_types = setup_dataloaders(
    #     config, mode=config.mode
    # )
    (
        model,
        # model_without_ddp,
        # optimizer,
        # scheduler,
        # scaler,
        tokenizer,
        # start_epoch,
        # global_step,
    ) = setup_model(
        config,
        model_cls=InternVideo2_Stage2,
        add_decoder=False,
        pretrain=True,
        # find_unused_parameters=False,
        device=device
    )
    model = model.to(device)
    model.eval()

    return model_path, model, tokenizer, config, create_transforms(224)


@torch.no_grad()
def evaluation(
    texts,
    image_paths,
    transforms,
    model,
    tokenizer,
    device,
    num_frames=4,
    max_txt_l=40,
):
    model.eval()

    text_feats, text_atts = extract_text_feats(
        texts, max_txt_l, tokenizer, model, device
    )  # (bsz, Lt, d), (bsz, Lt)

    image_feats, pooled_image_feats = extract_vision_feats(
        image_paths, transforms, model, device, num_frames=num_frames
    )  # (bsz, 1, #frm*Li, d) or (bsz, #frm, Li, d), (bsz, #frm, d)
    pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)
    i2t_scores, _ = get_sim(
        model.vision_proj(pooled_image_feats), model.text_proj(text_feats[:, 0])
    )

    text_encoder = model.get_text_encoder()
    encoder_output = image_feats[:, 0].to(device, non_blocking=True)
    encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(device, non_blocking=True)
    output = text_encoder(
        encoder_embeds=text_feats,
        attention_mask=text_atts,
        encoder_hidden_states=encoder_output,
        encoder_attention_mask=encoder_att,
        return_dict=True,
        mode="fusion",
    )

    itm_outputs = output.last_hidden_state[:, 0]
    itm_embeds = torch.cat([itm_outputs], dim=0)
    itm_scores = model.itm_head(itm_embeds)[:, 1]

    return (
        itm_scores,
        i2t_scores.diagonal()
    )



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn

from transformers import AutoConfig, AutoModelForCausalLM, \
                         LlamaConfig, LlamaModel, LlamaForCausalLM

from transformers.modeling_outputs import CausalLMOutputWithPast
from transformers.generation.utils import GenerateOutput

from ..llava_arch import LlavaMetaModel, LlavaMetaForCausalLM


class LlavaConfig(LlamaConfig):
    model_type = "llava_llama"


class LlavaLlamaModel(LlavaMetaModel, LlamaModel):
    config_class = LlavaConfig

    def __init__(self, config: LlamaConfig):
        super(LlavaLlamaModel, self).__init__(config)


class LlavaLlamaForCausalLM(LlamaForCausalLM, LlavaMetaForCausalLM):
    config_class = LlavaConfig

    def __init__(self, config):
        super(LlamaForCausalLM, self).__init__(config)
        self.model = LlavaLlamaModel(config)
        self.pretraining_tp = config.pretraining_tp
        self.vocab_size = config.vocab_size
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_model(self):
        return self.model

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[torch.FloatTensor] = None,
        image_sizes: Optional[List[List[int]]] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithPast]:

        if inputs_embeds is None:
            (
                input_ids,
                position_ids,
                attention_mask,
                past_key_values,
                inputs_embeds,
                labels
            ) = self.prepare_inputs_labels_for_multimodal(
                input_ids,
                position_ids,
                attention_mask,
                past_key_values,
                labels,
                images,
                image_sizes
            )

        return super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            labels=labels,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict
        )

    @torch.no_grad()
    def generate(
        self,
        inputs: Optional[torch.Tensor] = None,
        images: Optional[torch.Tensor] = None,
        image_sizes: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Union[GenerateOutput, torch.LongTensor]:
        position_ids = kwargs.pop("position_ids", None)
        attention_mask = kwargs.pop("attention_mask", None)
        if "inputs_embeds" in kwargs:
            raise NotImplementedError("`inputs_embeds` is not supported")

        if images is not None:
            (
                inputs,
                position_ids,
                attention_mask,
                _,
                inputs_embeds,
                _
            ) = self.prepare_inputs_labels_for_multimodal(
                inputs,
                position_ids,
                attention_mask,
                None,
                None,
                images,
                image_sizes=image_sizes
            )
        else:
            inputs_embeds = self.get_model().embed_tokens(inputs)

        return super().generate(
            position_ids=position_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            **kwargs
        )

    def prepare_inputs_for_generation(self, input_ids, past_key_values=None,
                                      inputs_embeds=None, **kwargs):
        images = kwargs.pop("images", None)
        image_sizes = kwargs.pop("image_sizes", None)
        inputs = super().prepare_inputs_for_generation(
            input_ids, past_key_values=past_key_values, inputs_embeds=inputs_embeds, **kwargs
        )
        if images is not None:
            inputs['images'] = images
        if image_sizes is not None:
            inputs['image_sizes'] = image_sizes
        return inputs

AutoConfig.register("llava_llama", LlavaConfig)
AutoModelForCausalLM.register(LlavaConfig, LlavaLlamaForCausalLM)


import logging
from contextlib import suppress

import torch
import torch.nn.functional as F
from open_clip import image_to_device
from tqdm import tqdm


def evaluate(model, dataloader, tokenizer, device, amp=True, args=None):
    """
    Evaluate the model on the given dataset.
    The task has N instances, each instance has I images and C captions.
    For each instance, the goal is to find the correct image for each caption and the correct caption for each image.
    This is done by computing the similarities between each image and each caption.
    This procedure is used to evaluate the models on Winoground and SugarCrepe.

    Parameters
    ----------

    model: torch.nn,Module
        CLIP-like model with `encode_image` and `encode_text`

    dataloader: torch.utils.data.Dataloader
        dataloader to use for evaluation

    tokenizer:
        text tokenizer, i.e. convert list of strings to torch.Tensor of integers

    device: cpu/cuda

    amp: whether to use automatic mixed precision

    Returns
    -------

    dict of accuracy metrics
    """
    autocast = torch.cuda.amp.autocast if amp else suppress
    image_score = []
    text_score = []
    score = []
    for batch_images, batch_texts in tqdm(dataloader):
        # assert(len(batch_images.shape) == 4)
        batch_images = image_to_device(
            batch_images,
            device,
            torch.float32,
            mean=args.image_mean,
            std=args.image_std,
        )
        # Because of the packing collate function we cannot support multi-image to caption selection
        nim = 1

        # tokenize all texts in the batch
        nt = len(batch_texts[0])
        batch_texts_tok_ = tokenizer(
            [text for i, texts in enumerate(batch_texts) for text in texts]
        ).to(device)

        # compute the embedding of images and texts
        with torch.no_grad(), autocast():
            batch_images_emb = F.normalize(
                model.encode_image(batch_images), dim=-1
            ).unsqueeze(1)
            B, _, emb_dim = batch_images_emb.shape
            batch_texts_emb = F.normalize(
                model.encode_text(batch_texts_tok_), dim=-1
            ).view(B, nt, -1)

        gt = torch.arange(min(nim, nt)).to(device)
        for i in range(B):
            # iteratve over instances

            # compute similarities between each image and each text
            images_emb = batch_images_emb[i]
            texts_emb = batch_texts_emb[i]
            scores = images_emb @ texts_emb.t()

            # i-th image should be matched to the i-th text
            image_closest_text = scores.argmax(dim=1)[: len(gt)]
            text_closest_image = scores.argmax(dim=0)[: len(gt)]
            pred_text_is_correct = (image_closest_text == gt).all().item()
            pred_image_is_correct = (text_closest_image == gt).all().item()
            all_correct = pred_text_is_correct and pred_image_is_correct
            image_score.append(pred_image_is_correct)
            text_score.append(pred_text_is_correct)
            score.append(all_correct)
    metrics = {}
    metrics["image_acc"] = torch.Tensor(image_score).float().mean().item()
    metrics["text_acc"] = torch.Tensor(text_score).float().mean().item()
    metrics["acc"] = torch.Tensor(score).float().mean().item()
    return metrics



import torch
import torch.nn as nn

from transformers import CLIPVisionModel, CLIPImageProcessor, CLIPVisionConfig


class CLIPVisionTower(nn.Module):
    def __init__(self, vision_tower, args, delay_load=False):
        super().__init__()

        self.is_loaded = False

        self.vision_tower_name = vision_tower
        self.select_layer = args.mm_vision_select_layer
        self.select_feature = getattr(args, 'mm_vision_select_feature', 'patch')

        if not delay_load:
            self.load_model()
        else:
            self.cfg_only = CLIPVisionConfig.from_pretrained(self.vision_tower_name)

    def load_model(self):
        self.image_processor = CLIPImageProcessor.from_pretrained(self.vision_tower_name)
        self.vision_tower = CLIPVisionModel.from_pretrained(self.vision_tower_name)
        self.vision_tower.requires_grad_(False)

        self.is_loaded = True

    def feature_select(self, image_forward_outs):
        image_features = image_forward_outs.hidden_states[self.select_layer]
        if self.select_feature == 'patch':
            image_features = image_features[:, 1:]
        elif self.select_feature == 'cls_patch':
            image_features = image_features
        else:
            raise ValueError(f'Unexpected select feature: {self.select_feature}')
        return image_features

    @torch.no_grad()
    def forward(self, images):
        if type(images) is list:
            image_features = []
            for image in images:
                image_forward_out = self.vision_tower(image.to(device=self.device, dtype=self.dtype).unsqueeze(0), output_hidden_states=True)
                image_feature = self.feature_select(image_forward_out).to(image.dtype)
                image_features.append(image_feature)
        else:
            image_forward_outs = self.vision_tower(images.to(device=self.device, dtype=self.dtype), output_hidden_states=True)
            image_features = self.feature_select(image_forward_outs).to(images.dtype)

        return image_features

    @property
    def dummy_feature(self):
        return torch.zeros(1, self.hidden_size, device=self.device, dtype=self.dtype)

    @property
    def dtype(self):
        return self.vision_tower.dtype

    @property
    def device(self):
        return self.vision_tower.device

    @property
    def config(self):
        if self.is_loaded:
            return self.vision_tower.config
        else:
            return self.cfg_only

    @property
    def hidden_size(self):
        return self.config.hidden_size

    @property
    def num_patches(self):
        return (self.config.image_size // self.config.patch_size) ** 2



import torch
import torch.distributed as dist
from ..utils.distributed import get_rank, is_dist_avail_and_initialized, is_main_process
import random
import logging

logger = logging.getLogger(__name__)


class MetaLoader(object):
    """ wraps multiple data loader """
    def __init__(self, name2loader):
        """Iterates over multiple dataloaders, it ensures all processes
        work on data from the same dataloader. This loader will end when
        the shorter dataloader raises StopIteration exception.

        loaders: Dict, {name: dataloader}
        """
        self.name2loader = name2loader
        self.name2iter = {name: iter(l) for name, l in name2loader.items()}
        name2index = {name: idx for idx, (name, l) in enumerate(name2loader.items())}
        index2name = {v: k for k, v in name2index.items()}

        iter_order = []
        for n, l in name2loader.items():
            iter_order.extend([name2index[n]]*len(l))

        random.shuffle(iter_order)
        iter_order = torch.Tensor(iter_order).to(torch.device("cuda")).to(torch.uint8)

        # sync
        if is_dist_avail_and_initialized():
            # make sure all processes have the same order so that
            # each step they will have data from the same loader
            dist.broadcast(iter_order, src=0)
        self.iter_order = [index2name[int(e.item())] for e in iter_order.cpu()]

        logger.info(str(self))

    def __str__(self):
        output = [f"MetaLoader has {len(self.name2loader)} dataloaders, {len(self)} batches in total"]
        for idx, (name, loader) in enumerate(self.name2loader.items()):
            output.append(
                f"dataloader index={idx} name={name}, batch-size={loader.batch_size} length(#batches)={len(loader)} "
            )
        return "\n".join(output)

    def __len__(self):
        return len(self.iter_order)

    def __iter__(self):
        """ this iterator will run indefinitely """
        for name in self.iter_order:
            _iter = self.name2iter[name]
            batch = next(_iter)
            yield name, batch


class MetaLoader_rs(object):
    """ wraps multiple data loader """
    def __init__(self, name2loader, skip_num=0):
        """Iterates over multiple dataloaders, it ensures all processes
        work on data from the same dataloader. This loader will end when
        the shorter dataloader raises StopIteration exception.

        loaders: Dict, {name: dataloader}
        """
        self.name2loader = name2loader
        name2index = {name: idx for idx, (name, l) in enumerate(name2loader.items())}
        index2name = {v: k for k, v in name2index.items()}

        iter_order = []
        for n, l in name2loader.items():
            iter_order.extend([name2index[n]]*len(l))

        random.shuffle(iter_order)
        iter_order = torch.Tensor(iter_order).to(torch.device("cuda")).to(torch.uint8)

        # sync
        if is_dist_avail_and_initialized():
            # make sure all processes have the same order so that
            # each step they will have data from the same loader
            dist.broadcast(iter_order, src=0)
        
        if skip_num > 0:
            iter_order_skip = iter_order[:skip_num]
            for k, v in index2name.items():
                media_step = (iter_order_skip == k).sum().item()
                name2loader[v].sampler.set_start_iter(media_step)
                logger.info(f"{v} dataloder skip steps: {media_step}")
            iter_order = iter_order[skip_num:]
            self.name2loader = name2loader
        else:
            logger.info("Do not skip steps for any dataloader!")
            for k, v in index2name.items():
                name2loader[v].sampler.set_start_iter(0)
                
        self.name2iter = {name: iter(l) for name, l in name2loader.items()}
        self.iter_idx = iter_order
        self.iter_order = [index2name[int(e.item())] for e in iter_order.cpu()]
        
        logger.info(str(self))

    def __str__(self):
        output = [f"MetaLoader has {len(self.name2loader)} dataloaders, {len(self)} batches in total"]
        for idx, (name, loader) in enumerate(self.name2loader.items()):
            length = (self.iter_idx == idx).sum()
            output.append(
                f"dataloader index={idx} name={name}, batch-size={loader.batch_size} length(#batches)={length} "
            )
        return "\n".join(output)

    def __len__(self):
        return len(self.iter_order)

    def __iter__(self):
        """ this iterator will run indefinitely """
        for name in self.iter_order:
            _iter = self.name2iter[name]
            batch = next(_iter)
            yield name, batch


"""Tests for semantic_alignment module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_semantic_alignment_basics():
    from ayase.modules.semantic_alignment import SemanticAlignmentModule
    _test_module_basics(SemanticAlignmentModule, "semantic_alignment")

def test_semantic_alignment_video(video_sample):
    from ayase.modules.semantic_alignment import SemanticAlignmentModule
    video_sample.quality_metrics = QualityMetrics()
    m = SemanticAlignmentModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# Copyright (2024) Bytedance Ltd. and/or its affiliates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import openai
import json
import time
from typing import List
import os

azure_endpoint = os.getenv("AZURE_ENDPOINT", "")
azure_api_key = os.getenv("OPENAI_API_KEY", "")

azure_gpt4v_client = openai.AzureOpenAI(
    azure_endpoint=azure_endpoint,
    api_version="2023-07-01-preview",
    api_key=azure_api_key,
    timeout = 120, 
)
azure_gpt4_client = openai.AzureOpenAI(
    azure_endpoint=azure_endpoint,
    api_version="2023-07-01-preview",
    api_key=azure_api_key,
    timeout = 120,
)
client = openai.AzureOpenAI(
    azure_endpoint=azure_endpoint,
    api_version="2023-07-01-preview",
    api_key=azure_api_key,
    timeout = 120,
)

# Function to encode the image
def encode_image(image_path):
  with open(image_path, "rb") as image_file:
    return base64.b64encode(image_file.read()).decode('utf-8')

def call_gemini_api(prompt: str, image_paths: List[str]=None, images_bs64: List[str]=None):
    assert image_paths is not None or images_bs64 is not None, "image_paths and images_bs64 cannot be both None."
    if images_bs64 is None:
        encoded_images = [encode_image(p) for p in image_paths]
    else:
        encoded_images = images_bs64
    completion = client.chat.completions.create(
        model="gemini-1.5-pro-preview", 
        messages=[
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": prompt,
                    },
                    *[{
                        "type": "image_url",
                        "image_url": {
                            "url": image
                        }
                    } for image in encoded_images],
                ]
            }
        ],
    )   
    return json.loads(completion.model_dump_json())['choices'][0]['message']['content']

def call_azure_gpt4v_api(prompt: str, image_paths: List[str]=None, images_bs64: List[str]=None):
    assert image_paths is not None or images_bs64 is not None, "image_paths and images_bs64 cannot be both None."
    if images_bs64 is None:
        encoded_images = [encode_image(p) for p in image_paths]
    else:
        encoded_images = images_bs64
    completion = azure_gpt4v_client.chat.completions.create(
        model="gptv", # or gpt-4o-2024-05-13
        messages=[
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": prompt
                    },
                    *[{
                        "type": "image_url",
                        "image_url": {
                            "url": f'data:image/png;base64,{image}',
                            "detail": "high"
                        }
                    } for image in encoded_images],
                ]
            }
        ],
        temperature = 0.7,
        top_p = 0.95,
        # max_tokens = 4096,
    )
    return json.loads(completion.model_dump_json())['choices'][0]['message']['content']
        
def call_azure_gpt_api(prompt: str, model = 'gpt-35-turbo'):
    
    completion = azure_gpt4_client.chat.completions.create(
        model=model,
        messages=[
            {
                "role": "user",
                "content": prompt
            }
        ]
    )
    return json.loads(completion.model_dump_json())['choices'][0]['message']['content']

retry_exceptions = [
    "qpm limit, you can apply for expansion on the platform",
    "reach token limit, you can apply for expansion on the platform",
    "Request timed out",
    "The service is temporarily unable to process your request.",
    "upstream failed to respond",
    "502 Bad Gateway",
]

def try_call_api(model, prompt: str, image_paths: List[str]=None, images_bs64: List[str]=None):
    q_success = False
    while q_success != True:
        try:
            if model == 'gptv':
                assert image_paths is not None or images_bs64 is not None, "image_paths and images_bs64 cannot be both None."
                gpt_q = call_azure_gpt4v_api(prompt, image_paths, images_bs64)
            elif model in ['gpt-4-1106-preview', 'gpt-35-turbo']:
                gpt_q = call_azure_gpt_api(prompt, model)
            else:
                raise ValueError(f'{model} is invalid.')
            q_success = True
            return gpt_q, 0
        except Exception as e:
            e = f'ERROR from try_call_api: {e}'
            print(e)
            hit = False
            for x in retry_exceptions:
                if x in e:
                    hit = True
                    time.sleep(10)
            if not hit:
                return e, 1
 
if __name__ == '__main__':
    resp = call_azure_gpt4v_api(prompt='Describe the image in detail.', image_paths=[os.path.dirname(os.path.abspath(__file__)) + '/../assets/figures/tarsier_logo.jpg'])



"""Stereoscopic Quality module.

Evaluates stereo 3D video comfort and quality:

  stereo_comfort_score — 0-100 (higher = more comfortable viewing)

For side-by-side (SBS) or top-bottom (TB) stereo content:
  1. Splits the frame into left/right views.
  2. Computes disparity map via stereo matching.
  3. Checks depth comfort zone (disparity range).
  4. Detects stereoscopic window violations.
  5. Estimates cross-talk via view similarity.

For standard 2D content: returns None (metric is inapplicable).
"""

import logging
from pathlib import Path
from typing import Optional, Tuple

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class StereoscopicQualityModule(PipelineModule):
    name = "stereoscopic_quality"
    description = "Stereo 3D comfort and quality assessment"
    default_config = {
        "stereo_format": "auto",  # "sbs" (side-by-side), "tb" (top-bottom), "auto"
        "subsample": 10,
        "max_frames": 30,
        "max_disparity_percent": 3.0,  # Max % of width for comfortable viewing
        "warning_threshold": 50.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.stereo_format = self.config.get("stereo_format", "auto")
        self.subsample = self.config.get("subsample", 10)
        self.max_frames = self.config.get("max_frames", 30)
        self.max_disp_pct = self.config.get("max_disparity_percent", 3.0)
        self.warning_threshold = self.config.get("warning_threshold", 50.0)

    def setup(self) -> None:
        logger.info("Stereoscopic quality: OpenCV stereo matching initialised")

    # ------------------------------------------------------------------
    # Format detection & view splitting
    # ------------------------------------------------------------------

    @staticmethod
    def _detect_stereo_format(frame: np.ndarray) -> Optional[str]:
        """Guess stereo format from frame aspect ratio.

        SBS: width ~2x height.  TB: height ~2x width.
        Standard 16:9 or similar: not stereo.
        """
        h, w = frame.shape[:2]
        aspect = w / max(h, 1)

        if aspect > 3.2:  # Very wide — likely SBS
            return "sbs"
        if aspect < 0.7:  # Very tall — likely TB
            return "tb"
        # Could be half-SBS (normal aspect but each half is squeezed)
        # For now, assume not stereo
        return None

    @staticmethod
    def _split_views(
        frame: np.ndarray, fmt: str
    ) -> Tuple[np.ndarray, np.ndarray]:
        h, w = frame.shape[:2]
        if fmt == "sbs":
            left = frame[:, : w // 2]
            right = frame[:, w // 2:]
        elif fmt == "tb":
            left = frame[: h // 2, :]
            right = frame[h // 2:, :]
        else:
            left = frame
            right = frame
        return left, right

    # ------------------------------------------------------------------
    # Stereo analysis
    # ------------------------------------------------------------------

    @staticmethod
    def _compute_disparity(left_gray: np.ndarray, right_gray: np.ndarray) -> np.ndarray:
        """Compute disparity map using semi-global block matching."""
        stereo = cv2.StereoSGBM_create(
            minDisparity=0,
            numDisparities=64,
            blockSize=5,
            P1=8 * 5 * 5,
            P2=32 * 5 * 5,
            disp12MaxDiff=1,
            uniquenessRatio=10,
            speckleWindowSize=100,
            speckleRange=32,
        )
        disp = stereo.compute(left_gray, right_gray).astype(np.float32) / 16.0
        return disp

    def _evaluate_stereo(self, left: np.ndarray, right: np.ndarray) -> float:
        """Score stereo quality for one frame pair.  Returns 0-100."""
        left_gray = cv2.cvtColor(left, cv2.COLOR_BGR2GRAY)
        right_gray = cv2.cvtColor(right, cv2.COLOR_BGR2GRAY)

        # Resize to match if needed
        h = min(left_gray.shape[0], right_gray.shape[0])
        w = min(left_gray.shape[1], right_gray.shape[1])
        left_gray = cv2.resize(left_gray, (w, h))
        right_gray = cv2.resize(right_gray, (w, h))

        disp = self._compute_disparity(left_gray, right_gray)

        # 1. Depth comfort: disparity range within comfortable limits
        valid = disp > 0
        if valid.sum() < 100:
            return 50.0

        disp_range = float(np.percentile(disp[valid], 95) - np.percentile(disp[valid], 5))
        max_comfortable = w * self.max_disp_pct / 100.0
        comfort = min(max_comfortable / max(disp_range, 1e-6), 1.0)

        # 2. Completeness: fraction of valid disparity pixels
        completeness = float(valid.sum()) / max(valid.size, 1)

        # 3. Smoothness: disparity gradient magnitude (less blocky = better)
        disp_dx = np.abs(np.diff(disp, axis=1))
        disp_dy = np.abs(np.diff(disp, axis=0))
        smoothness = 1.0 - min(float(disp_dx.mean() + disp_dy.mean()) / 10.0, 1.0)

        # 4. Cross-talk estimate: views should differ but not too much
        view_diff = cv2.absdiff(left_gray, right_gray)
        mean_diff = float(view_diff.mean()) / 255.0
        # Too similar (~0) = no stereo, too different (>0.3) = cross-talk/error
        crosstalk = 1.0 - abs(mean_diff - 0.05) * 10.0
        crosstalk = max(0, min(1, crosstalk))

        score = (
            0.35 * comfort
            + 0.25 * completeness
            + 0.20 * smoothness
            + 0.20 * crosstalk
        ) * 100

        return float(np.clip(score, 0, 100))

    # ------------------------------------------------------------------
    # Process
    # ------------------------------------------------------------------

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        try:
            score = self._process_video(sample.path)
            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.stereo_comfort_score = score

            if score < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low stereo comfort: {score:.1f}/100",
                        details={"stereo_comfort_score": score},
                        recommendation="Stereo depth may cause viewer discomfort.",
                    )
                )

            logger.debug(f"Stereo comfort for {sample.path.name}: {score:.1f}")

        except Exception as e:
            logger.error(f"Stereoscopic quality failed for {sample.path}: {e}")

        return sample

    def _process_video(self, path: Path) -> Optional[float]:
        cap = cv2.VideoCapture(str(path))
        if not cap.isOpened():
            return None

        # Read first frame to detect format
        ret, first_frame = cap.read()
        if not ret:
            cap.release()
            return None

        fmt = self.stereo_format
        if fmt == "auto":
            fmt = self._detect_stereo_format(first_frame)
            if fmt is None:
                cap.release()
                return None  # Not stereo content

        scores = []
        left, right = self._split_views(first_frame, fmt)
        scores.append(self._evaluate_stereo(left, right))

        idx = 1
        while len(scores) < self.max_frames:
            ret, frame = cap.read()
            if not ret:
                break
            if idx % self.subsample == 0:
                left, right = self._split_views(frame, fmt)
                scores.append(self._evaluate_stereo(left, right))
            idx += 1

        cap.release()
        return float(np.mean(scores)) if scores else None



from configs.data import *
from configs.model import *

# ========================= data ==========================
train_corpus = "webvid_debug"
train_file = "${available_corpus[${train_corpus}]}"  # for lazy evaluation
test_file = dict(ret_test=available_corpus["msrvtt_1k_test"])
test_types = ["ret_test"]
num_workers = 12

stop_key = None

# ========================= input ==========================
num_frames = 8
num_frames_test = 8
batch_size = 256
batch_size_test = 64
max_txt_l = 32

inputs = dict(
    image_res=224,
    video_input=dict(
        num_frames="${num_frames}",
        sample_type="rand",
        num_frames_test="${num_frames_test}",
        sample_type_test="middle",
        random_aug=False,
    ),
    max_txt_l=dict(image="${max_txt_l}", video="${max_txt_l}"),
    batch_size=dict(image="${batch_size}", video="${batch_size}"),
    batch_size_test=dict(image="${batch_size_test}", video="${batch_size_test}"),
)

# ========================= model ==========================
model = dict(
    model_cls="InternVideo2_CLIP_small",
    vision_encoder=dict(
        name="internvideo2",
        in_chans=3,
        patch_size=14,
        img_size=224,
        qkv_bias=False,
        drop_path_rate=0.,
        head_drop_path_rate=0.,
        embed_dim=768,
        num_heads=12,
        mlp_ratio=4,
        init_values=0.1,
        qk_normalization=True,
        depth=12,
        use_flash_attn=True,
        use_fused_rmsnorm=True,
        use_fused_mlp=True,
        fused_mlp_heuristic=1,
        drop_cls_token=False,
        attn_pool_num_heads=16,
        clip_embed_dim=768,
        layerscale_no_force_fp32=True,
        num_frames=8,
        tubelet_size=1,
        sep_pos_embed=False,
        use_checkpoint=False,
        checkpoint_num=0,
        align_dim=512,
    ),
    text_encoder=dict(
        name="mobileclip_b"
    ),
    temp=1 / 100.0,
    temp_min=1 / 100.0,
    freeze_vision=True,
    open_vision_clip_projector=True,
    freeze_text=True,
    open_text_projection=False,
    vision_ckpt_path="your_model_path/B14_dist_1B_stage2.pth",
    load_vision_ckpt_from_internvideo2_stage2=False,
    text_ckpt_path="your_model_path/mobileclip_blt.pt",
)

criterion = dict(
    loss_weight=dict(
        vtc=1.0, 
    ),  # 0: disabled.
)

optimizer = dict(
    opt="adamW",
    lr=4e-4,
    opt_betas=[0.9, 0.98],  # default
    weight_decay=0.2,
    max_grad_norm=-1,  # requires a positive float, use -1 to disable
    # use a different lr for some modules, e.g., larger lr for new modules
    different_lr=dict(enable=False, module_names=[], lr=1e-3),
)

scheduler = dict(sched="cosine", epochs=3, min_lr_multi=0.01, warmup_epochs=0.6)

evaluate = True
deep_fusion = False
evaluation = dict(
    eval_frame_ensemble="concat",  # [concat, max, mean, lse]
    eval_x_only=False,
    k_test=128,
    eval_offload=True,  # offload gpu tensors to cpu to save memory.
)

use_half_precision = True
use_bf16 = True
gradient_checkpointing = True

# ========================= wandb ==========================
wandb = dict(
    enable=False,
    entity="likunchang",  # username or team name to store the runs, see https://docs.wandb.ai/ref/python/init
    project="InternVideo2_CLIP",  # setup in your command line
)
dist_url = "env://"
device = "cuda"
mode = "pt"

# ========================= others ==========================
output_dir = None  # output dir
resume = False  # if True, load optimizer and scheduler states as well
debug = False
log_freq = 1
seed = 42

save_latest = False
save_iter = 500
auto_resume = True
pretrained_path = ""  # path to pretrained model weights, for resume only?

deepspeed = dict(
    enable=False,
    stage=0,
)


# Copyright (c) Meta Platforms, Inc. and affiliates.

import os
from logging import getLogger
from typing import Callable, List, Tuple

import torch
from PIL import Image
from torchcodec.decoders import VideoDecoder
from torchvision.utils import draw_bounding_boxes

from ...core.transforms.image_transform import ImageTransform

logger = getLogger()


def get_video_transform(
    image_res: int = 224,
    normalize_img: bool = True,
) -> Tuple[Callable, int]:

    transforms = VideoTransform(
        size=image_res,
        normalize_img=normalize_img,
    )

    return transforms


class VideoTransform(ImageTransform):
    def __init__(
        self,
        size: int = 224,
        normalize_img: bool = True,
    ) -> None:
        super().__init__(
            size=size,
            normalize_img=normalize_img,
        )

    def __call__(self, video_info: tuple, sampling_fps: int = 1):
        video_path, max_frames, s, e, bbox_map = video_info

        frames, sample_pos = self.load_video(
            video_path,
            max_frames=max_frames,
            sampling_fps=sampling_fps,
            s=s,
            e=e,
        )

        if bbox_map:
            bbox_dict_map = {}
            for idx_pos, pos in enumerate(sample_pos):
                if str(pos) in bbox_map and bbox_map[str(pos)] is not None:
                    bbox_dict_map[idx_pos] = bbox_map[str(pos)]
            if len(bbox_dict_map) > 0:
                frames = self.draw_bounding_boxes(frames, bbox_dict_map)

        return super()._transform_torch_tensor(frames)

    def _process_multiple_images(self, image_paths: List[str]):
        images = [Image.open(path).convert("RGB") for path in image_paths]
        processed_images = []
        for image in images:
            image, (w, h) = super().__call__(image)
            processed_images.append(image)
        processed_images = torch.cat(processed_images, dim=0)
        return processed_images, (w, h)

    def _process_multiple_images_pil(self, images: List[Image.Image]):
        processed_images = []
        for image in images:
            image, (w, h) = super().__call__(image)
            processed_images.append(image)
        processed_images = torch.cat(processed_images, dim=0)
        return processed_images, (w, h)

    def load_video(self, video_path, max_frames=16, sampling_fps=1, s=None, e=None):
        """
        Loads a video from a given path and extracts frames based on specified parameters using OpenCV.

        Args:
            video_path (str): The path to the video file.
            max_frames (int, optional): The maximum number of frames to extract. Defaults to 16.
            sampling_fps (int, optional): The sampling frame rate. Defaults to 1.
            s (float, optional): The start time of the video in seconds. Defaults to None.
            e (float, optional): The end time of the video in seconds. Defaults to None.

        Returns:
            list: A list of frames extracted from the video.
        """

        if not os.path.exists(video_path):
            return

        decoder = VideoDecoder(video_path, device="cpu")
        decoder_metadata = decoder.metadata
        fps = decoder_metadata.average_fps
        total_frames = decoder_metadata.num_frames

        start_frame = 0 if s is None else int(s * fps)
        end_frame = total_frames - 1 if e is None else int(e * fps)
        end_frame = min(end_frame, total_frames - 1)

        if start_frame > end_frame:
            start_frame, end_frame = end_frame, start_frame
        elif start_frame == end_frame:
            end_frame = start_frame + 1

        sample_fps = int(sampling_fps)
        t_stride = int(round(float(fps) / sample_fps))

        all_pos = list(range(start_frame, end_frame + 1, t_stride))
        if len(all_pos) > max_frames:
            sample_idxs = self.uniform_sample(len(all_pos), max_frames)
            sample_pos = [all_pos[i] for i in sample_idxs]
        elif len(all_pos) < max_frames:
            total_clip_frames = end_frame - start_frame + 1
            if total_clip_frames < max_frames:
                max_frames = total_clip_frames
            sample_idxs = self.uniform_sample(total_clip_frames, max_frames)
            sample_pos = [start_frame + idx for idx in sample_idxs]
        else:
            sample_pos = all_pos

        all_frames = decoder.get_frames_at(indices=sample_pos)
        all_frames = all_frames.data

        return all_frames, sample_pos

    def uniform_sample(self, m, n):
        assert n <= m
        stride = (m - 1) / (n - 1) if n > 1 else 0  # Calculate the stride
        return [int(round(i * stride)) for i in range(n)]

    def draw_bounding_boxes(self, frames, all_bboxes):
        # Assuming frames is a torch.Tensor with shape (N, C, H, W)
        N, _, _, _ = frames.shape
        frames_with_bbox = (
            frames.clone()
        )  # Clone the tensor to avoid modifying the original
        for i in range(N):
            if i in all_bboxes:
                bbox = all_bboxes[i]
                # Convert bbox to a tensor and add a batch dimension
                bbox_tensor = torch.tensor([bbox], dtype=torch.float32)
                # Draw the bounding box on the frame
                frames_with_bbox[i] = draw_bounding_boxes(
                    frames_with_bbox[i],
                    boxes=bbox_tensor,
                    colors=(255, 0, 0),  # Red color for the bounding box
                    width=4,
                )
        return frames_with_bbox



import time
from functools import partial, reduce

import torch
import torch.nn as nn
from torch.nn.functional import adaptive_avg_pool3d

from .conv_backbone import convnext_3d_small, convnext_3d_tiny, convnextv2_3d_pico, convnextv2_3d_femto
from .head import IQAHead, VARHead, VQAHead
from .swin_backbone import SwinTransformer2D as ImageBackbone
from .swin_backbone import SwinTransformer3D as VideoBackbone
from .swin_backbone import swin_3d_small, swin_3d_tiny


class BaseEvaluator(nn.Module):
    def __init__(
        self, backbone=dict(), vqa_head=dict(),
    ):
        super().__init__()
        self.backbone = VideoBackbone(**backbone)
        self.vqa_head = VQAHead(**vqa_head)

    def forward(self, vclip, inference=True, **kwargs):
        if inference:
            self.eval()
            with torch.no_grad():
                feat = self.backbone(vclip)
                score = self.vqa_head(feat)
            self.train()
            return score
        else:
            feat = self.backbone(vclip)
            score = self.vqa_head(feat)
            return score

    def forward_with_attention(self, vclip):
        self.eval()
        with torch.no_grad():
            feat, avg_attns = self.backbone(vclip, require_attn=True)
            score = self.vqa_head(feat)
            return score, avg_attns


class DOVER(nn.Module):
    def __init__(
        self,
        backbone_size="divided",
        backbone_preserve_keys="fragments,resize",
        multi=False,
        layer=-1,
        backbone=dict(
            resize={"window_size": (4, 4, 4)}, fragments={"window_size": (4, 4, 4)}
        ),
        divide_head=False,
        vqa_head=dict(in_channels=768),
        var=False,
    ):
        self.backbone_preserve_keys = backbone_preserve_keys.split(",")
        self.multi = multi
        self.layer = layer
        super().__init__()
        for key, hypers in backbone.items():
            print(backbone_size)
            if key not in self.backbone_preserve_keys:
                continue
            if backbone_size == "divided":
                t_backbone_size = hypers["type"]
            else:
                t_backbone_size = backbone_size
            if t_backbone_size == "swin_tiny":
                b = swin_3d_tiny(**backbone[key])
            elif t_backbone_size == "swin_tiny_grpb":
                # to reproduce fast-vqa
                b = VideoBackbone()
            elif t_backbone_size == "swin_tiny_grpb_m":
                # to reproduce fast-vqa-m
                b = VideoBackbone(window_size=(4, 4, 4), frag_biases=[0, 0, 0, 0])
            elif t_backbone_size == "swin_small":
                b = swin_3d_small(**backbone[key])
            elif t_backbone_size == "conv_tiny":
                b = convnext_3d_tiny(pretrained=False)
            elif t_backbone_size == "conv_small":
                b = convnext_3d_small(pretrained=False)
            elif t_backbone_size == "conv_femto":
                b = convnextv2_3d_femto(pretrained=True)
            elif t_backbone_size == "conv_pico":
                b = convnextv2_3d_pico(pretrained=True)
            elif t_backbone_size == "xclip":
                raise NotImplementedError
                # b = build_x_clip_model(**backbone[key])
            else:
                raise NotImplementedError
            print("Setting backbone:", key + "_backbone")
            setattr(self, key + "_backbone", b)
        if divide_head:
            for key in backbone:
                pre_pool = False #if key == "technical" else True
                if key not in self.backbone_preserve_keys:
                    continue
                b = VQAHead(pre_pool=pre_pool, **vqa_head)
                print("Setting head:", key + "_head")
                setattr(self, key + "_head", b)
        else:
            if var:
                self.vqa_head = VARHead(**vqa_head)
                print(b)
            else:
                self.vqa_head = VQAHead(**vqa_head)

    def forward(
        self,
        vclips,
        inference=True,
        return_pooled_feats=False,
        return_raw_feats=False,
        reduce_scores=False,
        pooled=False,
        **kwargs
    ):
        assert (return_pooled_feats & return_raw_feats) == False, "Please only choose one kind of features to return"
        if inference:
            self.eval()
            with torch.no_grad():
                scores = []
                feats = {}
                for key in vclips:
                    feat = getattr(self, key.split("_")[0] + "_backbone")(
                        vclips[key], multi=self.multi, layer=self.layer, **kwargs
                    )
                    if hasattr(self, key.split("_")[0] + "_head"):
                        scores += [getattr(self, key.split("_")[0] + "_head")(feat)]
                    else:
                        scores += [getattr(self, "vqa_head")(feat)]
                    if return_pooled_feats:
                        feats[key] = feat
                    if return_raw_feats:
                        feats[key] = feat
                if reduce_scores:
                    if len(scores) > 1:
                        scores = reduce(lambda x, y: x + y, scores)
                    else:
                        scores = scores[0]
                    if pooled:
                        scores = torch.mean(scores, (1, 2, 3, 4))
            self.train()
            if return_pooled_feats or return_raw_feats:
                return scores, feats
            return scores
        else:
            self.train()
            scores = []
            feats = {}
            for key in vclips:
                feat = getattr(self, key.split("_")[0] + "_backbone")(
                    vclips[key], multi=self.multi, layer=self.layer, **kwargs
                )
                if hasattr(self, key.split("_")[0] + "_head"):
                    scores += [getattr(self, key.split("_")[0] + "_head")(feat)]
                else:
                    scores += [getattr(self, "vqa_head")(feat)]
                if return_pooled_feats:
                    feats[key] = feat.mean((-3, -2, -1))
            if reduce_scores:
                if len(scores) > 1:
                    scores = reduce(lambda x, y: x + y, scores)
                else:
                    scores = scores[0]
                if pooled:
                    print(scores.shape)
                    scores = torch.mean(scores, (1, 2, 3, 4))
                    print(scores.shape)

            if return_pooled_feats:
                return scores, feats
            return scores
        
    def forward_head(
        self,
        feats,
        inference=True,
        reduce_scores=False,
        pooled=False,
        **kwargs
    ):
        if inference:
            self.eval()
            with torch.no_grad():
                scores = []
                feats = {}
                for key in feats:
                    feat = feats[key]
                    if hasattr(self, key.split("_")[0] + "_head"):
                        scores += [getattr(self, key.split("_")[0] + "_head")(feat)]
                    else:
                        scores += [getattr(self, "vqa_head")(feat)]
                if reduce_scores:
                    if len(scores) > 1:
                        scores = reduce(lambda x, y: x + y, scores)
                    else:
                        scores = scores[0]
                    if pooled:
                        scores = torch.mean(scores, (1, 2, 3, 4))
            self.train()
            return scores
        else:
            self.train()
            scores = []
            feats = {}
            for key in vclips:
                feat = getattr(self, key.split("_")[0] + "_backbone")(
                    vclips[key], multi=self.multi, layer=self.layer, **kwargs
                )
                if hasattr(self, key.split("_")[0] + "_head"):
                    scores += [getattr(self, key.split("_")[0] + "_head")(feat)]
                else:
                    scores += [getattr(self, "vqa_head")(feat)]
                if return_pooled_feats:
                    feats[key] = feat
            if reduce_scores:
                if len(scores) > 1:
                    scores = reduce(lambda x, y: x + y, scores)
                else:
                    scores = scores[0]
                if pooled:
                    print(scores.shape)
                    scores = torch.mean(scores, (1, 2, 3, 4))
                    print(scores.shape)

            if return_pooled_feats:
                return scores, feats
            return scores
        
class MinimumDOVER(nn.Module):
    def __init__(self):
        super().__init__()
        self.technical_backbone = VideoBackbone()
        self.aesthetic_backbone = convnext_3d_tiny(pretrained=False)
        self.technical_head = VQAHead(pre_pool=False, in_channels=768)
        self.aesthetic_head = VQAHead(pre_pool=False, in_channels=768)


    def forward(self,aesthetic_view, technical_view):
        self.eval()
        with torch.no_grad():
            aesthetic_score = self.aesthetic_head(self.aesthetic_backbone(aesthetic_view))
            technical_score = self.technical_head(self.technical_backbone(technical_view))
            
        aesthetic_score_pooled = torch.mean(aesthetic_score, (1,2,3,4))
        technical_score_pooled = torch.mean(technical_score, (1,2,3,4))
        return [aesthetic_score_pooled, technical_score_pooled]



class BaseImageEvaluator(nn.Module):
    def __init__(
        self, backbone=dict(), iqa_head=dict(),
    ):
        super().__init__()
        self.backbone = ImageBackbone(**backbone)
        self.iqa_head = IQAHead(**iqa_head)

    def forward(self, image, inference=True, **kwargs):
        if inference:
            self.eval()
            with torch.no_grad():
                feat = self.backbone(image)
                score = self.iqa_head(feat)
            self.train()
            return score
        else:
            feat = self.backbone(image)
            score = self.iqa_head(feat)
            return score

    def forward_with_attention(self, image):
        self.eval()
        with torch.no_grad():
            feat, avg_attns = self.backbone(image, require_attn=True)
            score = self.iqa_head(feat)
            return score, avg_attns



from abc import abstractmethod
from typing import List
import torch

from ..model import ScoreModel

class VQAScoreModel(ScoreModel):

    @abstractmethod
    def forward(self,
                images: List[str],
                texts: List[str],
                question_template: str,
                answer_template: str) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        question_template: a string with optional {} to be replaced with the 'text'
        answer_template: a string with optional {} to be replaced with the 'text'
        """
        pass